
# The lib (data) directory in structure.
# todo: rewrite basic name patterns (define beyond \w.-/)
from os.path import join as joinpath, sep as PATHSEP
from glob import glob
import re
import yaml

class LibDataError:
    pass

class LibDataUriFormatError(LibDataError, ValueError):
    pass
class LibDataUnknownUriSection(LibDataError, NameError):
    pass
class LibDataUnknownUriName(LibDataError, NameError):
    pass

def getFullSection(section, subsection = None):
    if subsection:
        return '%s[%s]' % (section, subsection)

    return section

def getMiddleNameKey(name):
    c = name[:1].lower()
    if c in 'abcde': return 'A-E'
    if c in 'fghij': return 'F-J'
    if c in 'klmno': return 'K-O'
    if c in 'pqrst': return 'P-T'
    if c in 'uvwxyz': return 'U-Z'
    return 'ZZZ'

def merge(d, d2):
    d.update(d2)
    return d

def group(i):
    if isinstance(i, dict):
        i = iter(i.items())

    r = dict()
    for (n, v) in i:
        r.setdefault(n, []).append(v)

    return r

def getUnderAssumedRoot(*parts):
    parts = PATHSEP.join(parts).split(PATHSEP)
    path = []

    for p in parts:
        if p == '..':
            if path:
                path.pop()

        elif p != '.':
            path.append(p)

    return PATHSEP.join(path)

def joinUnderAssumedRoot(root, *parts):
    return joinpath(root, getUnderAssumedRoot(*parts))

class Section(dict):
    def __init__(self, section, subsection = None):
        self.section = section.upper()
        self.subsection = subsection.upper() if subsection else None

    def getFullSection(self):
        return getFullSection(self.section, self.subsection)
    def getLeafpoint(self, name):
        name = name.lower()
        try: return self[name]
        except KeyError:
            raise LibDataUnknownUriName(name)

    # leaf iteration?

    def enumerate(self):
        return iter(self.items())

    def toString(self):
        return yaml.dump(self.toStructure())
    def toStructure(self):
        return dict(section = self.getFullSection(), mapping = dict(self))

    __str__ = toString

# Section Types.
class PathMap(Section):
    def __init__(self, section, path, *components, **mapping):
        Section.__init__(self, section)
        self.path = path

        dict.__init__(self, mapping)

        for name in components:
            self[name] = name

    def getLeafpoint(self, name):
        return joinUnderAssumedRoot(self.path, Section.getLeafpoint(self, name))
    def toStructure(self):
        return dict(section = self.section, path = self.path, mapping = dict(self))

    def enumerate(self):
        from os.path import join
        path = self.path

        for (name, mapping) in self.items():
            yield (name, join(path, mapping))

    @classmethod
    def FromTree(self, section, tree):
        try: mapping = tree['mapping']
        except KeyError: mapping = {}

        components = tree.get('components', [])
        return self(section, tree['path'], *components, **mapping)

class RecordMap(Section):
    def __init__(self, section, varname, format, glob):
        Section.__init__(self, section)
        self.format = format # validate?
        self.varname = varname
        self.glob = glob

    def getLeafpoint(self, name):
        # Ugh, what about case?
        return self.format % {self.varname: getUnderAssumedRoot(name)}

    def enumerate(self):
        if self.glob is None:
            raise ValueError('Glob pattern not configured')

        for i in glob(self.glob):
            # todo: translate the glob into a regex pattern,
            # such that it can be used to match each entry 'i'
            # for the name that identifies the entry, and use
            # it as the 'name' item for this result.
            yield (None, i)

    def toStructure(self):
        return dict(section = self.section, format = self.format,
                    variable = self.varname)

    @classmethod
    def FromTree(self, section, tree):
        return self(section, tree.get('variable', 'name'),
                    tree['format'], tree.get('glob'))

class WorldRecord(RecordMap):
    def __init__(self, section, varname, format, index_format):
        RecordMap.__init__(self, section, varname, format, None)
        self.index_format = index_format

    def getLeafpoint(self, name):
        if name.isdigit():
            return RecordMap.getLeafpoint(self, name)

        name = name.lower()
        if name not in ['index', 'index.mini']:
            raise LibDataUnknownUriName(name)

        return self.index_format % dict(index = name)

    def enumerate(self, index = None):
        if index is None:
            index = 'index'

        indexEntry = self.getLeafpoint(index)
        yield (index, indexEntry)

        for n in open(indexEntry):
            # Yes, read the index file.
            n = n.strip()
            # xxx should be parsing *.suffix
            if n.isdigit():
                yield (n, self.getLeafpoint(n))

    @classmethod
    def FromTree(self, section, tree):
        return self(section, tree.get('variable', 'zone'),
                    tree['format'], tree['index-format'])

class PartitionMap(RecordMap):
    def __init__(self, *args, **kwd):
        RecordMap.__init__(self, *args, **kwd)
        self.section_varname = self.varname + '.section'

    def getLeafpoint(self, name):
        name = getUnderAssumedRoot(name)
        return self.format % {self.varname: name,
                              self.section_varname: getMiddleNameKey(name)}

    def toStructure(self):
        return dict(section = self.section, format = self.format,
                    variable = self.varname)

# Root.
class Directory(Section):
    def __init__(self, name, entries):
        Section.__init__(self, name)
        dict.__init__(self, ((e.getFullSection(), e) for e in entries))

    def toStructure(self):
        return dict((n, v.toStructure()) for (n, v) in self.items())

    SECTION_DEF_PATTERN = re.compile(r'([\w.]*)(?:\(([\w-]+)\))?')
    SECTION_TYPES = {'path-map': PathMap,
                     'record-map': RecordMap,
                     'partition-map': PartitionMap,
                     'world-record': WorldRecord}

    @classmethod
    def FromTree(self, name, tree):
        if isinstance(tree, str):
            tree = yaml.load(tree)

        def extractSections(subtree, parent = None):
            for (name, config) in subtree.items():
                m = self.SECTION_DEF_PATTERN.match(name)
                assert m is not None

                (name, type) = m.groups()
                if type is None:
                    # Define a parent section.
                    if name.endswith('[]'):
                        name = name[:-2]

                    for s in extractSections(config, parent = name):
                        yield s
                else:
                    typeClass = self.SECTION_TYPES[type]
                    section = name

                    if parent:
                        if not section:
                            # The default space for this section parent.
                            section = parent
                        else:
                            section = '%s[%s]' % (parent, section)

                    yield typeClass.FromTree(section, config)

        return self(name, extractSections(tree))

    # Todo: rewrite alot of this.
    # Accessors -- todo: get rid of whole notion of subtype in this parse: just keep it opaque.
    URI_SECTIONSUB_PATTERN = R'(?P<section>\w+)(?:\[(?P<subsection>[^]]+)\])?'
    URI_PATTERN = r'^(?:stuph\:)?%s\:(?P<name>[\w./]+)$' % URI_SECTIONSUB_PATTERN
    URI_SECTIONSUB_PATTERN = re.compile(URI_SECTIONSUB_PATTERN)
    URI_PATTERN = re.compile(URI_PATTERN)

    @classmethod
    def parse(self, uri):
        m = self.URI_PATTERN.match(uri)
        if m is None:
            raise LibDataUriFormatError(uri)

        m = m.groupdict()
        section = m['section'].upper()
        subsection = m['subsection']
        name = m['name']

        if subsection:
            section = '%s[%s]' % (section, subsection.upper())

        return (section, name)

    def lookup(self, name):
        try: section = dict.__getitem__(self, name)
        except KeyError:
            raise LibDataUnknownUriSection(name)

        return section

    def lookupFilename(self, uri):
        (section, name) = self.parse(uri)
        return self.lookup(section).getLeafpoint(name)

    __getitem__ = lookupFilename

    @property
    def contents(self):
        return Contents(self)


class Contents: # Archival Operation
    # Kind of a heavy-weight class because it reconstructs alot of the work
    # done to tear down the directory structure..  But it's a directly
    # functional implementation.
    #
    # Access the iterator of toplevel sectional structures, and then
    # iterate through those by specifying an (world) index, enumerating
    # each entry mapping for a handle to the resource content.

    def __init__(self, directory):
        self.directory = directory

    MAIN_SECTION_ENUMERATION_ORDER = []

    def __iter__(self):
        def gathering():
            # Scan sections, parse structure and group.
            for (_, section) in self.directory.enumerate():
                (name, resource) = self.directory.parse \
                                  ('%s:NULL' % section.getFullSection())

                m = self.directory.URI_SECTIONSUB_PATTERN.match(name)
                if m is None:
                    print(name) # because it shouldn't fail
                    yield (name, (None, section))
                else:
                    (name, sub) = m.groups()
                    # if sub is None:
                    #     print name, repr(section)[:20]

                    yield (name, (sub, section))

        groups = iter(group(gathering()).items())

        ordering = self.MAIN_SECTION_ENUMERATION_ORDER.index
        def mainSectionOrderFor(xxx_todo_changeme):
            (name, item) = xxx_todo_changeme
            try: return ordering(name)
            except ValueError:
                return (0,) # last, ascending

        groups = sorted(groups, key = mainSectionOrderFor)

        # XXX PLAYER[MAIL(.{COUNT})]
        # This can be done with specific subs.
        # It could also be done by re-encapsulating the partition-map type.
        # XXX group
        return (self.Section.FromGroup(*group)
                for group in groups)

    class Section(list):
        # Structured internal representation of directory configuration.
        @classmethod
        def FromGroup(self, name, group):
            group = dict(group) # item pairs of sub/object
            main = group.pop(None, None)

            return self(self.MakeSub(None, name, main),
                        *(self.MakeSub(name, *v) for v in
                          group.items()))

        def __repr__(self):
            return '%s: %d subsections' % (self.main.name,
                                           len(self))

        @classmethod
        def MakeSub(self, root, name, object):
            root = '%s_%s' % (root, name.replace('.', '_')) \
                   if root else name

            subClass = self._subClasses.get(root, self.Sub)
            return subClass(name, object)

        def __init__(self, main, *subs):
            list.__init__(self, subs)
            self.main = main

        def __iter__(self):
            yield self.main

            for sub in self.subSections:
                yield sub

        @property
        def subSections(self):
            return list.__iter__(self)

        class Sub:
            # A sub-section (or, a section's main node).
            def __init__(self, name, object):
                self.name = name
                self.object = object

            def __repr__(self):
                return '%s: %r' % (self.name, self.object)

            @property
            def hasContents(self):
                return self.object is not None

            def __getitem__(self, index):
                if isinstance(self.object, WorldRecord):
                    return self.object.enumerate(index)

                return self.object.enumerate()

            def __iter__(self):
                # Prevent python from trying to access getitem sequentially.
                raise NotImplementedError


        ##    class WorldSub(Sub):
        ##        def __getitem__(self, index):
        ##            return self.object.enumerate(index)

        ##    _subClasses = dict(PLAYER = Player,
        ##                       PLAYER_MAIL = PlayerMail,
        ##                       PLAYER_MAIL_COUNT = PlayerMailCount)

        _subClasses = dict()

    def contents(self, index = None):
        # Generates a sequence of individual file components
        # paired with a qualified section key as well as a
        # unique value that may identify this particular
        # component resource within that directory section.
        #
        # Also references a particular index into each section,
        # which reflects the selection of library entities
        # that should be included in this content.

        def _(key, s):
            if s.hasContents:
                for c in s[index]:
                    yield (key, c)

        for section in self:
            main = section.main
            mainName = main.name

            for c in _((mainName, None), main):
                yield c

            for sub in section.subSections:
                key = (mainName, sub.name)

                for c in _(key, sub):
                    yield c

    __call__ = contents
    mainIndex = property(contents)

    def __str__(self):
        # XXX nls, mapi, group, iteritems
        return nls('%s:\n%s\n' % \
                   (n[0] if n[1] is None else '%s[%s]' % \
                    (n[0], n[1]), indent(nls(sorted(mapi(sliceOf[1], v)))))
                   for (n, v) in iteritems(group(self.mainIndex)))


# Program the Main Directory -- todo: build with yaml.
MAIN = Directory.FromTree('MAIN',
        '''
        GAME:
            (path-map):
                path: etc
                components: ['bank', 'dns.cache', 'hcontrol', 'olczones',
                             'players', 'shopgold', 'time', 'trust',
                             'trust.python']

                mapping:
                    network.sessions: network-sessions.bin
                    process.id: process_id
                    syslog.patterns: syslog-patterns.cfg
                    shell.commands: shell-commands.cfg
                    admin.script: admin_scripts/
                    plugins: plugins/
                    deployment: deployment/
                    config: config.cfg
                    zones: zone-modules.cfg
                    subdaemons: subdaemons.cfg

            TEXT(path-map):
                path: text
                components: ['background', 'clanhelp', 'credits', 'handbook', 'immlist',
                             'imotd', 'info', 'motd', 'news', 'policies', 'wizlist']

            MISC(path-map):
                path: misc
                components: ['bugs', 'ideas', 'messages', 'socials', 'typos', 'xnames']

        CLAN:
            (path-map):
                path: clan
                components: ['bank', 'index']

            BAN(record-map):
                variable: clan
                format: clan/ban/%(clan)s.ban
                glob: clan/ban/*.ban
            DATA(record-map):
                variable: clan
                format: clan/data/%(clan)s.dat
                glob: clan/data/*.dat
            INFO(record-map):
                variable: clan
                format: clan/info/%(clan)s.inf
                glob: clan/info/*.inf

        BOARD(record-map):
            variable: name
            format: boards/board.%(name)s
            glob: boards/board.*

        HELP:
            (path-map):
                path: text/help
                components: ['screen']

            TOPIC(record-map):
                variable: topic
                format: text/help/%(topic)s

                # XXX Should be what's in the index!
                glob: text/help/{index,general,skills,wizhelp,programmer}

        WORLD:
            ZONE(world-record):
                format: world/zon/%(zone)s.zon
                index-format: world/zon/%(index)s
            ROOM(world-record):
                format: world/wld/%(zone)s.wld
                index-format: world/wld/%(index)s
            ITEM(world-record):
                format: world/obj/%(zone)s.obj
                index-format: world/obj/%(index)s
            MOBILE(world-record):
                format: world/mob/%(zone)s.mob
                index-format: world/mob/%(index)s

            SHOP(world-record):
                format: world/shp/%(zone)s.shp
                index-format: world/shp/%(index)s
            SEARCH(world-record):
                format: world/sch/%(zone)s.sch
                index-format: world/sch/%(index)s

            SPECIALS(partition-map):
                format: world/specials/%(name)s.spc
                glob: world/specials/*.spc

            HOUSE(partition-map):
                variable: house
                format: house/%(house)s.house
                glob: house/*.house

        PLAYER:
            RENT(partition-map):
                format: plrobjs/%(name.section)s/%(name)s.objs
                glob: plrobjs/{A-E,F-J,K-O,P-T,U-Z,ZZZ}/*.objs
            STRINGS(partition-map):
                format: plrstrings/%(name.section)s/%(name)s.strings
                glob: plrstrings/{A-E,F-J,K-O,P-T,U-Z,ZZZ}/*.strings
            ALIAS(partition-map):
                format: plralias/%(name.section)s/%(name)s.alias
                glob: plralias/{A-E,F-J,K-O,P-T,U-Z,ZZZ}.alias

            MAIL(partition-map):
                format: plrmail/%(name.section)s/%(name)s.mail
                glob: plrmail/{A-E,F-J,K-O,P-T,U-Z,ZZZ}.mail
            MAIL.COUNT(partition-map):
                format: plrmail/%(name.section)s/%(name)s.count
                glob: plrmail/{A-E,F-J,K-O,P-T,U-Z,ZZZ}.count
        ''')

StuphLIB = MAIN


# Directory Accessors.
getLibraryFile = MAIN.lookupFilename

def openLibraryFile(uri, interfaceRequest):
    # Try to open it for editing (with interface).
    pass

def pathNormalize(path):
    r = []
    b = ''

    for c in path:
        if c in '/\\':
            if b:
                r.append(b)
                b = ''

        else:
            b += c

    if b:
        r.append(b)

    return r

def doIncrementFilename(filename):
    # Essentially, create only if not exists,
    # reserving by way of this algorithm (but not locking).
    # XXX O_CREATONLY
    touch(filename)
    return filename

def getIncrementedFilename(filename):
    i = filename.find('.')
    if i < 0:
        root = filename
        ext = ''
    else:
        root = filename[:i]
        ext = filename[i:]

    while True:
        if not exists(filename):
            return doIncrementFilename(filename)

        parts = root.split('/')
        last = parts[-1] # assert

        if last.isdigit():
            last = int(last)
            parts = parts[:-1]
            parts.append(str(last + 1))
        else:
            parts.append('2')

        filename = '-'.join(parts) + ext


def defaultDateFormatter0(timestamp):
    pass

class DateFormatter:
    def __init__(self, format):
        self.format = format
    def __call__(self, timestamp):
        pass # return timestamp.strptime(self.format)

def compileDateFormat(format):
    if format == DEFAULT_DATE_FORMAT0:
        return defaultDateFormatter0

    return DateFormatter(format)

def getEntriesFromArchivalOptions(options):
    root = options.archive_root
    root = pathNormalize(root) if root else []

    index = options.world_index
    sectionize = options.archive_sections

    for ((section, sub), (key, entry)) \
        in MAIN.contents(index):

        # MAIN['SECTION[SUB]:KEY'] -> open(entry)
        localpath = entry

        entry = pathNormalize(entry)
        path = list(root)

        if sectionize:
            path.append(section.lower())
            if sub:
                path.append(sub.lower())

        path.extend(entry)
        path = '/'.join(path)

        yield (localpath, path)

archivalOptions = getEntriesFromArchivalOptions

def doZipArchive(filename, options, entries):
    # XXX memory hog:
    # todo: utilize shell utilities and disk space
    from zipfile import ZipFile
    zf = ZipFile(filename, 'w')

    for (localpath, path) in entries:
        # todo: a map-file-to-string
        try: o = open(localpath)
        except IOError as e:
            print(e)
            continue

        contents = o.read()
        zf.writestr(path, contents)

    zf.close()
    return zf

def doBZip2TarArchive(filename, options, entries):
    raise NotImplementedError
def doGZipTarArchive(filename, options, entries):
    raise NotImplementedError
def doDirectoryStructure(filename, options, entries):
    raise NotImplementedError
def doRDBMS(filename, options, entries):
    raise NotImplementedError

##    class doZipArchive(object):
##        __new__ = staticmethod(doZipArchive0)
##
##        @classmethod
##        def InitializeCmdlnOptions(self, parser):
##            parser.add_option('--use-shell', action = 'store_true')


def getConfigOptions(resource, section = None):
    # Read the configuration file specified in the directory.
    # return io.path(MAIN[CONFIG_RESOURCE_FORMAT % \
    #                dict(resource = resource)]) \
    #           .loading.ini

    return dict()


DEFAULT_OUTPUT_TARGET = 'zip-archive'

DEFAULT_DATE_FORMAT0 = 'mmmddthyyyy'
DEFAULT_DATE_FORMAT = DEFAULT_DATE_FORMAT0

OUTPUT_TARGET_ALIASES = dict(zip = 'zip-archive',
                             bz = 'bzip-tar',
                             bz2 = 'bzip-tar',
                             bzip = 'bzip-tar',
                             bzip2 = 'bzip-tar',
                             gz = 'gzip-tar',
                             gzip = 'gzip-tar',
                             folder = 'directory-structure')

OUTPUT_TARGETS = {'zip-archive': doZipArchive,
                  'bzip-tar': doBZip2TarArchive,
                  'gzip-tar': doGZipTarArchive,
                  'directory-structure': doDirectoryStructure,
                  'rdbms': doRDBMS}

CONFIG_RESOURCE_FORMAT = 'GAME:%(resource)s' # 'GAME[CONFIG]:%(resource)s'
DEFAULT_CONFIG_RESOURCE = 'config'
DEFAULT_CONFIG_SECTION = ''

def doArchiveContents(filename, options):
    # PYTHONPATH="..\..\pythonlib\packages"
    # python -m stuphlib.directory
    #   --archive=stuphlib-aug18th2014.zip

    outputTarget = options.output_target
    outputTarget = OUTPUT_TARGET_ALIASES.get \
                   (outputTarget, outputTarget)

    try: outputTarget = OUTPUT_TARGETS[outputTarget]
    except KeyError:
        raise NameError('%s not in available %s' % \
                        (outputTarget, list(OUTPUT_TARGETS.keys())))


    dateFormatter = compileDateFormat(options.date_format)
    entries = archivalOptions(options)
    configOptions = getConfigOptions(options.config_resource)

    from datetime import datetime

    todaysDate = dateFormatter(datetime.now())

    filename = filename.format(dict(today = todaysDate,
                                    game = configOptions))

    if options.increment:
        filename = doIncrementFilename(filename)

    return outputTarget(filename, options, entries)


def main(argv = None):
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('--dump')
    parser.add_option('--dump-full', action = 'store_true')

    parser.add_option('--archive-contents', '--archive')
    parser.add_option('--output-target', default = DEFAULT_OUTPUT_TARGET)
    parser.add_option('--increment', action = 'store_true')
    parser.add_option('--date-format', default = DEFAULT_DATE_FORMAT)

    parser.add_option('--config-resource', default = DEFAULT_CONFIG_RESOURCE)
    parser.add_option('--config-section', default = DEFAULT_CONFIG_SECTION)

    parser.add_option('--archive-root')
    parser.add_option('--archive-sections', action = 'store_true')

    parser.add_option('--world-index')

    (options, args) = parser.parse_args(argv)

    if options.archive_contents:
        doArchiveContents(options.archive_contents, options)
    else:
        if options.dump_full:
            print(MAIN.toString())
        if options.dump:
            print(MAIN.lookup(options.dump).toString())

        for a in args:
            print(MAIN[a])

if __name__ == '__main__':
    main()

# Reference.
'''
etc/
    bank|dns.cache|hcontrol|olczones|players|shopgold|time|trust

    config.cfg
    zone-modules.cfg
    trust.python
    shell-commands.cfg
    subdaemons.cfg
    syslog-patterns.cfg

    network-sessions.bin
    process_id

    admin_scripts/
    plugins/

clan/
    ban/
        number.ban
    data/
        number.dat
    inf/
        number.inf

boards/
    board.name

text/
    background|clanhelp|credits|handbook|immlist|imotd
    info|motd|news|policies|wizlist

text/help/
    screen
    general|general|wizhelp|skills|programmer

misc/
    bugs|ideas|messages|socials|typos|xnames

world/
    mob|obj|sch|shp|wld|zon
        number.section
        index
        index.mini

    specials/
        name.spc
        zone-modules.cfg

plrobjs/
    A-E|F-J|K-O|P-T|U-Z|ZZZ/
        name.objs

plrstrings/
    A-E|F-J|K-O|P-T|U-Z|ZZZ/
        name.strings

plralias/
    A-E|F-J|K-O|P-T|U-Z|ZZZ/
        name.alias

plrmail/
    A-E|F-J|K-O|P-T|U-Z|ZZZ/
        name.mail
        name.count

house/
    number.house
'''

'''
stuphlib::
    boards: board.*
    house: *.house

    text:
        background, credits, handbook, immlist, imotd, info, motd,
        news, policies, title, wizlist, clanhelp

        help:
            screen, general, skills, wizhelp, index, programmer

    misc:
        bugs, ideas, typos, socials, messages, xnames, treeMap

    clan:
        bank, index

        ban: *.ban
        data: *.dat
        info: *.inf

    etc:
        hcontrol, olczones, shopgold, bank, time, trust, players,
        dns.cache

    plrstrings:
        {A-E,F-J,K-O,P-T,U-Z,ZZZ}/*.strings
    plrobjs:
        {A-E,F-J,K-O,P-T,U-Z,ZZZ}/*.objs
    plrmail:
        {A-E,F-J,K-O,P-T,U-Z,ZZZ}/*.{mail,count}
    plralias:
        {A-E,F-J,K-O,P-T,U-Z,ZZZ}/*.alias

    world:
        zon: *.zon, index
        wld: *.wld, index
        obj: *.obj, index
        mob: *.mob, index
        shp: *.shp, index
        sch: *.sch, index
        '''



'Package designed for isolating the import call so that no relative modules block others in the toplevel.'

def LookupObject(name, raise_import_errors = False):
    'A clever routine that can import an object from a system module from any attribute depth.'
    # XXX This doesn't import modules that aren't imported by their parent package initialization code.  (Yes it does :skip:)
    parts = name.split('.')

    moduleObject = None
    module = None
    n = len(parts)
    x = 0

    importError = None

    for x in range(n, 0, -1):
        name = '.'.join(parts[:x])

        try: moduleObject = __import__(name, globals(), locals())
        except (ImportError, ModuleNotFoundError):
            from sys import exc_info
            importError = exc_info()
        else:
            break

    # while x < n:
    #     name = '.'.join(parts[:x+1])
    #     moduleObject = module

    #     # Try to find the module that contains the object.
    #     try: module = __import__(name, globals(), locals(), [''])
    #     except ImportError:
    #         from sys import exc_info
    #         importError = exc_info()
    #         break
    #     else:
    #         # Q: Will this break the following code?  It needs to be here so that a full
    #         # subpackage name can be looked up, when no sub-object is needed.  Because,
    #         # when x < n, the (module)Object is returned and this setlocal phase will be
    #         # skipped.
    #         moduleObject = module

    #     x += 1

    # No top-level module could be imported.
    if moduleObject is None:
        if importError is None:
            raise AssertionError

        (etype, value, tb) = importError
        raise value.with_traceback(tb)

    object = moduleObject
    x = 1

    while x < n:
        # Now look for the object value in the module.

        # If an attribute can't be found -- this is where we raise the original import-error?
        #
        # This is a good idea: so that LookupObject is consistantly raising ImportError (since
        # that's basically what its function is -- an import-from).
        #
        ##    if raise_import_errors:
        ##        module = __import__(name, globals(), locals(), [''])

        try: object = getattr(object, parts[x])
        except AttributeError as e:
            if importError is None:
                raise e

            (c, e, t) = importError
            raise c(e).with_traceback(t)

        x += 1

    # todo: raise prior ImportError if it existed because of a problem in the loaded module.
    return object



from receptacle.application import Application, Configuration, StorageManagement, RightsManagement, ApiManagement

from stuphos.runtime.facilities import Facility
from ..etc import nth
from stuphos import getConfig

from .service import Girl as GirlService

class ReceptacleApplicationManager(Facility, Application):
    NAME = 'Receptacle::Application::Manager'

    def __init__(self):
        argv = []
        config_file = getConfig('configuration', 'Receptacle')
        config = Configuration.FromCmdln(argv, config_file,
                                         GirlService, StorageManagement,
                                         RightsManagement, ApiManagement,
                                         default_section = 'Application')

        Application.__init__(self, config)
        nth(self.run)




# MUD . Player . Interfaces
# todo: rename to mud.interface.shell
from stuphmud import server as mud
from . import dictcmd

class ACMD:
    Name = '?'
    def __init__(self, **kwd):
        self.__dict__.update(kwd)

class CommandCenter(dictcmd.CmdDict):
    # This can probably go away and just use CmdDict..?

    def __init__(self):
        dictcmd.CmdDict.__init__(self)
        self.setup()

    def setup(self):
        pass

    def importCommands(self, module):
        for proc in  __import__(module).__all__:
            if callable(proc) and hasattr(proc, 'ACMD'):
                self.insertAssignment(proc.ACMD.Name, proc)

def getCommandCenter(module):
    if not hasattr(module, 'SharedCommands'):
        from .standard import StandardCommands
        module.SharedCommands = StandardCommands()
        module.ACMD = module.assignCommand = module.SharedCommands.assign

    return module.SharedCommands

# Python implementation of a command manager that can execute verbs out of a dictionary.
class CommandManager:
    def __init__(self, commands = None):
        self.commands = commands

    # Fraun Dec 21st 2005 Incorporation of CmdDict
    def playerCommand(self, peer, cmd, argstr):
        # Fraun Oct 1st 2006 - Only playing connections for commands!
        if peer.avatar:
            # todo: mud.tasks.triggers.verbCommand

            action = self.commands.lookup(cmd)
            if callable(action):
                try: return action(peer, cmd, argstr)
                except (mud.player.PlayerResponse, mud.player.DoNotHere) as e:
                    print(e, file=peer) # e.message
                    return True

                except mud.player.PlayerAlert as e:
                    e.deliver(peer)
                    return True

    try: from world.commands import parseCommand
    except ImportError:
        def parseCommand(self, cmdln):
            return self.commands.parse(cmdln)

    def assignCommand(self, *args):
        return self.commands.insertOverridingAll(*args)

    def getCommands(self):
        return self.commands

from .history import HistoryManager
from .threading import ThreadingManager
from .code import CodeManager, makeCodeEvaluators



# Copyleft Clint Banis 2006
# Note the hack in PLRFILE.read_player_entries and read_player
'''
DEFAULT_PLRFL = 'lib/etc/players'
DEFAULT_STORE = 'lib/etc/players.db' # ':memory:'

PLRFILE().read_player_entries()
'''

DEFAULT_PLRFL = 'etc/players'
DEFAULT_STORE = 'etc/players.db' # ':memory:'

__all__=['PLRFILE', 'PLRDBIntoSQL', 'DEFAULT_PLRFL', 'DEFAULT_STORE', 'unpack_player', 'tuple2dict', 'cstring', 'trim_cstrings']

# import new

from struct import *
from os.path import exists
from pprint import pprint as pp

try: from stuphos.etc.tools.structfmt import stfmti
except ImportError:
	def stfmti(*args, **kwd):
		raise NotImplementedError

import stuphmud.server.player

# ToDo: incorporate the original pfile-position
# Composite unique key formed by idnum/name and save number/save-time

saving_throw_names=('savingthrow1', 'savingthrow2', 'savingthrow3', 'savingthrow4', 'savingthrow5')

char_specials_names=('alignment', 'idnum', 'act', 'affectedby') +saving_throw_names
player_specials_names=('skills', 'talks', 'wimp', 'freeze', 'invis', 'loadroom',
				'pref', 'pref2', 'disp', 'badpws', 'drunk', 'hunger', 'thirst',
				'spells2learn', 'maritalstatus', 'num_marriages', 'spouse',
				'spellattack1', 'spellattack2', 'remortpoints')

abilities_names=('str', 'stradd', 'intel', 'wis', 'dex', 'con', 'cha')
points_names=('mana', 'maxmana', 'hit','maxhit', 'move', 'maxmove', 'deaths', 'mkills', 'pkills',
		'dts', 'qpoints', 'armor', 'gold', 'bank', 'exp', 'hitroll', 'damroll')

pfilel_names=('name', 'email', 'remort', 'race', 'clan', 'clanrank', 'pagelength', 'saveroom',
		'sex', 'chclass', 'level', 'hometown', 'birth', 'played', 'weight', 'height', 'pwd')\
		+char_specials_names +player_specials_names +abilities_names +points_names\
		+('affected', 'color', 'lastlogin', 'lastlogon', 'host')

#pfilel_types=('varchar(%d)'%MAX_NAME_LENGTH+1, 'varchar(%d)'%MAX_EMAIL_LENGTH+1)+('integer',)*6

pfilel_ncount=len(pfilel_names)
pfilel_types=['']*pfilel_ncount

MAX_NAME_LENGTH=20
MAX_EMAIL_LENGTH=80
MAX_PWD_LENGTH=10
MAX_AFFECT=32
MAX_COLOR_SET=32
HOST_LENGTH=40
MAX_SKILLS=200
MAX_TONGUE=3

char_specials_fmt    = 'i3l5h'
player_specials_fmt  = '%ds' % (calcsize('b') * MAX_SKILLS+1)
player_specials_fmt += 'x'
player_specials_fmt += '%ds' % (calcsize('i') * MAX_TONGUE)
player_specials_fmt += 'ibhI3lB3b3il3i'
abilities_fmt        = '7b'
points_fmt           = '12h3i2b'

affected_fmt         = '2h2bl'
affected_fmt        += '%dx' % calcsize('P') # padd out next pointer

# ending_padding     = '1191x'
ending_padding       = ''

u_header_fmt         = '%ds%ds6i3bh2i2B%ds' % (MAX_NAME_LENGTH+1, MAX_EMAIL_LENGTH+1, MAX_PWD_LENGTH+1)
u_affected_fmt       = '%ds' % (calcsize(affected_fmt) * MAX_AFFECT)	# XXX Do not use string-type
u_color_fmt          = '%ds' % (calcsize('i') * MAX_COLOR_SET)		# XXX Do not use string-type
u_logtime_fmt        = '2i'
u_host_fmt           = '%ds' % (HOST_LENGTH+1)

pfilel_fmt           = u_header_fmt
pfilel_fmt          += char_specials_fmt
pfilel_fmt          += player_specials_fmt
pfilel_fmt          += abilities_fmt
pfilel_fmt          += points_fmt
pfilel_fmt          += u_affected_fmt
pfilel_fmt          += u_color_fmt
pfilel_fmt          += u_logtime_fmt
pfilel_fmt          += u_host_fmt

pfilel_fmt          += ending_padding

def printFormats(formats = ('u_header_fmt',
				'char_specials_fmt',
				'player_specials_fmt',
				'abilities_fmt',
				'points_fmt',
				'u_affected_fmt',
				'u_color_fmt',
				'u_logtime_fmt',
				'u_host_fmt')):

	tmpl = '%-20s %5d %5d %s'

	print('Name                  Size #Atms Format')
	print(tmpl % ('pfilel_fmt', calcsize(pfilel_fmt), len(list(stfmti(pfilel_fmt))), pfilel_fmt))
	
	total_size = 0
	z = 0

	for n in formats:
		f = globals()[n]
		s = calcsize(f)
		t = len(list(stfmti(f)))

		print(tmpl % (n, s, t, (' ' * z) + f))

		total_size += s
		z += len(f)

	# Total size should be the same as calcsize(pfilel_fmt)
	print('Total Size:', total_size)

def printNames(formats = ('pfilel_names',
				'saving_throw_names',
				'char_specials_names',
				'player_specials_names',
				'abilities_names',
				'points_names')):

	print('Name                  Size        Format')
	for n in formats:
		f = globals()[n]
		print('%-20s %5d       %r' % (n, len(f), f))

pfilel_size=calcsize(pfilel_fmt)
pfilel_sqldecl=', '.join([' '.join(p) for p in zip(pfilel_names, pfilel_types)])

NEW_PLAYERS_TABLE='CREATE TABLE players(%s)' % ', '.join(pfilel_names)
NEW_PLAYER='INSERT OR REPLACE INTO players('+pfilel_sqldecl+') VALUES('+', '.join(['?']*pfilel_ncount)+')'

def unpack_player(buf, cleanse_nonstring_buffers = True):
	t = unpack(pfilel_fmt, buf + '\x00') ## XXX Hack

	if cleanse_nonstring_buffers:
		# Kill the affected and colorset buffers because they are not UTF8 friendly.
		# This is dissimilar from trim_cstrings

		t = list(t)  # make mutable
		t[71] = ''   # position of affected buffer
		t[72] = ''   # position of colorset buffer
		# Also, talks, and skills are non-UTF8-friendly buffers
		t = tuple(t) # reformat

	return t

def load_player_stream(stream, unpack = False):
	# iapply
	read = stream.read

	while True:
		buf = read(pfilel_size - 1) ## XXX Hack
		if buf == '':
			return

		if unpack:
			yield unpack_player(buf)
		else:
			yield buf

def read_player_entries(file = None, filename = ''):
	if file:
		if type(file) is str:
			filename = file
		else:
			return load_player_stream(file)

	if filename:
		return load_player_stream(open(filename))

	raise AssertionError

def goto_pfilepos(fl, pfilepos):
	return fl.seek(pfilepos*pfilel_size)

def read_player(playerfl, pfilepos):
	goto_pfilepos(playerfl, pfilepos)
	return unpack_pack(playerfl.read(pfilel_size - 1) + '\x00')  ## XXX Hack

def write_player(playerfl, pfilepos, plrrecord):
	goto_pfilepos(playerfl, pfilepos)
	return playerfile.write(pack(pfilel_fmt, plrrecord))

## Aliases:
class PlayerFileReader(object):
	# XXX Backwards-compatability.
	def __init__(self, players = None):
		self.players = players or DEFAULT_PLRFL

	def __repr__(self):
		return 'players: ' + repr(self.players)

	def read_player_entries(self, filename = None):
		'Wrapper for contained playerfile.read_player_entries.'
		return read_player_entries(filename or self.players)

PLRFILEReader=PLRFILE=PlayerFileReader


class PLRDBIntoSQL(PlayerFileReader):
	"Plugs a PLRFILE into an SQLite database"

	def __init__(self, store=None, loadsource=None):
		# PlayerFileReader.__init__(self)
		self.open_store(type(store) is str and (store,) or store)
		if loadsource:
			self.save_players(self.read_player_entries(loadsource))

	def write_player(self, player, pfilepos = 0):
		'Install player into cursory connection'
		# pp(player)
		# raw_input()
		# print 'len(player[1])', len(player[1])

		self.x(NEW_PLAYER, player)
		# self.x('UPDATE players SET pfilepos=%d WHERE name=%r' % (pfilepos, player[1][0]))

	def save_players(self, players = None, ):
		"Stream all players given by the `players' iterable through write_player -- Also creates self.stmts and self.players."
		plrs=self.players=[]
		append=plrs.append

		stmts=self.stmts=[]
		append2=stmts.append

		pfilepos=0
		for p in players:
			append2(p[0])
			append(p[1])

			self.write_player(p, pfilepos)
			pfilepos+=1

		print('Saved', len(plrs), 'Players ...')

	def open_store(self, store):
		e=not exists(store[0])

		c=self.connection = mud.player.sqlite_connect(*store)
		u=self.u=c.cursor()
		x=self.x=u.execute

		if e:
			x(NEW_PLAYERS_TABLE)
			x('ALTER TABLE players ADD (pfilepos)')

	def commit(self):
		self.connection.commit()

def tuple2dict(t, names = pfilel_names):
	
	d = {}

	for (n, v) in zip(names, t):
		d[n] = v

	return d

def cstring(s):
	i = s.find('\x00')
	if i < 0:
		return s
	return s[:i]

def trim_cstrings(d):
	def reset(n):
		if n in d:
			d[n] = cstring(d[n])

	if type(d) is dict:
		reset('name')
		reset('email')
		reset('host')

	return d

if __name__=='__main__':
	from psyco import full;full()

	plrdb=PLRDBIntoSQL(DEFAULT_STORE)
	plrdb.save_players(plrdb.read_player_entries(DEFAULT_PLRFL))
	plrdb.commit()

	u=plrdb.u
	x=plrdb.x
	a=lambda:x('select * from players')



# World-Class Serialization.
from stuphos.etc import fromJsonFile, fromJsonString, toJsonFile, toJsonString

# Affected by equipments:
# ['aff_charisma', 'aff_constitution', 'aff_dexterity', 'aff_intelligence',
#  'aff_strength', 'aff_strength_addition', 'aff_wisdom']
# ['armorclass', 'damroll', 'hitroll']
# ['carry_items', 'carry_weight']
# ['breath', 'height', 'weight']

BASIC_MOBILE_DATA = ['alignment', 'attack_type', 'charisma', 'constitution',
                     'damnodice', 'damsizedice', 'default_pos', 'description', 'dexterity',
                     'experience', 'gold_on_hand', 'intelligence', 'last_dir',
                     'level', 'long_descr', 'name', 'position', 'sex', 'short_descr',
                     'strength', 'strength_addition', 'timer', 'track_timer', 'waitState',
                     'walk_type', 'was_in_room', 'wisdom', 'affections']

BASIC_ITEM_DATA =  ['detailed_description', 'exdescs', 'keywords', 'name',
                    'room_description', 'type', 'value1', 'value2', 'value3', 'value4']

VERSION = 1

class RealWorld:
    def __init__(self):
        from stuphos.system.api import syslog, world
        players = world.player.players
        zone = world.zone
        room = world.room
        item = world.item
        mobile = world.mobile
        iterate_entities = world.iterate_entities

        self.RealZone = zone
        self.RealRoom = room
        self.ItemPrototype = item
        self.MobilePrototype = mobile
        self.IterateEntities = iterate_entities
        self.PlayingConnections = players
        self.syslog = syslog

class WorldFreezer(RealWorld):
    CRITERIA = ['exits', 'combat', 'empty_zones', 'all']
    def __init__(self, options):
        RealWorld.__init__(self)
        self.version = VERSION

        # Initialize criteria.
        for n in self.CRITERIA:
            setattr(self, n, False)

        # Detect general criteria.
        if options.complete or options.world == 'complete':
            for n in self.CRITERIA:
                setattr(self, n, True)

        # Set specific criteria.
        elif options.world:
            for n in options.world.split(','):
                if n in self.CRITERIA:
                    setattr(self, n, True)

        self.item_ref = {}
        self.mobile_ref = {}
        self.player_ref = {}

    def nextItemId(self, item):
        try: return self.item_ref[item]
        except KeyError:
            nr = self.item_ref[item] = len(self.item_ref)
            return nr

    def nextMobileId(self, mobile):
        try: return self.mobile_ref[mobile]
        except KeyError:
            nr = len(self.mobile_ref)
            if mobile.isPlayer:
                nr = -nr

            self.mobile_ref[mobile] = nr
            return nr

    def getItemId(self, item):
        return self.item_ref.get(item)
    def getMobileId(self, mobile):
        return self.mobile_ref.get(mobile)

    def getPlayerData(self, player):
        nr = self.nextMobileId(player)
        try: return self.player_ref[nr]
        except KeyError:
            data = self.player_ref[nr] = {'idnum': player.idnum,
                                          'name': player.name}
            return data

    def flagSet(self, name, instance, prototype):
        instance = getattr(instance, name).set
        if prototype is None:
            return instance

        prototype = getattr(prototype, name).set
        return list(set(instance).difference(prototype))

    def doFlagSet(self, data, name, *args):
        value = self.flagSet(name, *args)
        if value:
            data[name] = value

    def doMaxValueSet(self, data, name, instance):
        maxname = 'max_%s' % name
        v = getattr(instance, name)
        max = getattr(instance, maxname)
        if v != max:
            data[name] = v
            data[maxname] = max

    ##    def doItemValueSet(self, data, item, prototype):
    ##        if prototype is not None:
    ##            for n in xrange(1, 5):
    ##                n = 'value%d' % n
    ##                v1 = getattr(item, n)
    ##                v2 = getattr(prototype, n)
    ##                if v1 != v2:
    ##                    data[n] = v1
    ##
    ##        else:
    ##            for n in xrange(1, 5):
    ##                n = 'value%d' % n
    ##                data[n] = getattr(item, n)

    def itemData(self, item):
        # Get basic item data.  Filter out prototypical-similarities.
        data = {}
        prototype = item.prototype
        for n in BASIC_ITEM_DATA:
            try:
                v = getattr(item, n)
                if getattr(prototype, n) != v:
                    data[n] = v

            except AttributeError:
                pass

        # Now add in more complex data.
        # self.doItemValueSet(data, item, prototype)

        self.doFlagSet(data, 'anticlass', item, prototype)
        self.doFlagSet(data, 'antiflags', item, prototype)
        self.doFlagSet(data, 'extraflags', item, prototype)
        self.doFlagSet(data, 'wearflags', item, prototype)

        contents = list(map(self.itemData, item.contents))
        if contents: data['contents'] = contents

        # writing (mail, note, etc.)

        if prototype is not None:
            data['vnum'] = item.vnum

        data['instance.id'] = self.nextItemId(item)
        return data

    def npcData(self, mobile):
        # Get basic mobile data. Filter out prototypical-similarities.
        data = {}
        prototype = mobile.prototype
        for n in BASIC_MOBILE_DATA:
            try:
                v = getattr(mobile, n)
                if getattr(prototype, n) != v:
                    data[n] = v

            except AttributeError:
                pass

        # Now add in more complex data.
        self.doFlagSet(data, 'affectflags', mobile, prototype)
        self.doFlagSet(data, 'npcflags', mobile, prototype)

        self.doMaxValueSet(data, 'fatigue', mobile)
        self.doMaxValueSet(data, 'hit', mobile)
        self.doMaxValueSet(data, 'mana', mobile)

        for n in ('opponent', 'master', 'holding', 'mount'):
            o = getattr(mobile, n)
            if o:
                data[n] = self.nextMobileId(o)

        inventory = list(map(self.itemData, mobile.inventory))
        if inventory: data['inventory'] = inventory

        if mobile.tracking >= 0:
            data['tracking'] = mobile.tracking

        # equipment

        # If any specific data is available, then generate new instance.id
        if data or self.all:
            # Vnum is always present, directly.
            data['vnum'] = mobile.vnum
            data['instance.id'] = self.nextMobileId(mobile)
            return data

    def playerData(self, player):
        o = player.opponent
        if o and self.combat:
            data = self.getPlayerData(player)
            data['opponent'] = self.nextMobileId(o)

    def mobileData(self, mobile):
        if mobile.npc:
            return self.npcData(mobile)

        return self.playerData(mobile)

    def isHouse(self, room):
        return bool(room.house) and room.flags.House

    BASIC_EXIT_STATE = set(['Closed', 'Locked'])

    def roomData(self, room):
        items = None
        if not self.isHouse(room):
            items = [i for i in (self.itemData(i) for i in room.contents) if i]

        exits = None
        if self.exits:
            exits = ((e, self.BASIC_EXIT_STATE.intersection(e.flags.set)) for e in room.exits)
            exits = dict((e.direction, list(flags)) for (e, flags) in exits if flags)

        mobiles = [m for m in (self.mobileData(m) for m in room.people) if m]
        if items or mobiles or exits:
            data = dict(vnum = room.vnum)
            if items:
                data['items'] = items
            if mobiles:
                data['mobiles'] = mobiles
            if exits:
                data['exits'] = exits

            return data

    def isFreezable(self, zone):
        # Internal definition of is_empty:
        #   Linked player (with descriptor)
        #   in CON_PLAYING state
        #   must be level 105 or lower
        return self.empty_zones or not zone.empty

    def zoneData(self, zone):
        if self.isFreezable(zone):
            data = dict(vnum = zone.vnum, age = zone.age)
            rooms = [r for r in (self.roomData(r) for r in zone.rooms) if r]
            if rooms:
                data['rooms'] = rooms

            return data

    def iterateZones(self):
        return self.IterateEntities(self.RealZone)

    def dump(self):
        from world import iterent, zone as ZoneType
        players = self.player_ref
        zones = [z for z in (self.zoneData(z) for z in self.iterateZones()) if z]

        data = dict(version = self.version)
        if players:
            data['players'] = players
        if zones:
            data['zones'] = zones

        return data

class WorldUnfreezer(RealWorld):
    def __init__(self):
        RealWorld.__init__(self)
        self.item_ref = {}
        self.mobile_ref = {}

    def doSetFlags(self, bitv, flags):
        if flags:
            for f in flags:
                setattr(bitv, f, True)

    def restoreItem(self, location, data):
        try: vnum = data['vnum']
        except KeyError:
            # XXX oldgen not implemented
            self.syslog("COPYOVER: OLDGEN Object not supported <%d>!" % data['instance.id'])
            return

        warned_of = set()
        item = self.ItemPrototype(vnum).new(location)
        for name in BASIC_ITEM_DATA:
            try: setattr(item, name, data[name])
            except KeyError:
                pass
            except AttributeError:
                # Note: the AttributeError is for non-writable binding members.
                if name not in warned_of:
                    warned_of.add(name)
                    self.syslog('COPYOVER: Non-Writable Mobile Attribute: %s' % name)

        for name in range(1, 5):
            name = 'value%d' % name

            try: setattr(item, name, data[name])
            except (KeyError, AttributeError):
                pass

        for name in ('anticlass', 'antiflags', 'extraflags', 'wearflags'):
            self.doSetFlags(getattr(item, name), data.get(name))

        self.item_ref[data['instance.id']] = (item, data)
        return item

    def restoreMobile(self, location, data):
        warned_of = set()
        mobile = self.MobilePrototype(data['vnum']).instantiate(location)
        for name in BASIC_MOBILE_DATA:
            try: setattr(mobile, name, data[name])
            except KeyError:
                pass
            except AttributeError:
                # Note: the AttributeError is for non-writable binding members.
                if name not in warned_of:
                    warned_of.add(name)
                    self.syslog('COPYOVER: Non-Writable Item Attribute: %s' % name)

        for name in ('affectflags', 'npcflags'):
            self.doSetFlags(getattr(mobile, name), data.get(name))

        for name in ('fatigue', 'hit', 'mana', 'max_fatigue', 'max_hit', 'max_mana'):
            try: setattr(mobile, name, data[name])
            except KeyError:
                pass
            except AttributeError:
                self.syslog('COPYOVER: Non-Writable Mobile Attribute: %s' % name)

        self.mobile_ref[data['instance.id']] = (mobile, data)
        return mobile

    def restoreRoom(self, room, data):
        # todo: traverse these things backwards...
        for item in data.get('items') or []:
            self.restoreItem(room, item)
        for mobile in data.get('mobiles') or []:
            self.restoreMobile(room, mobile)

        for (exit, state) in data.get('exits', {}).items():
            exit = getattr(room, exit)
            if exit is not None:
                self.doSetFlags(exit.flags, state)

    def findLoadedPlayerByIdnum(self, idnum):
        for player in self.PlayingConnections():
            if player.avatar and player.avatar.idnum == idnum:
                return player.avatar

    def restorePlayers(self, players):
        ref = {}
        if players:
            for (nr, data) in players.items():
                loaded = self.findLoadedPlayerByIdnum(data['idnum'])

                assert loaded
                assert loaded.name == data['name']
                ref[nr] = loaded

                try: loaded.opponent = self.mobile_ref[data['opponent']][0]
                except IndexError:
                    pass

        return ref

    def load(self, data):
        for zone in data.get('zones') or []:
            real_zone = self.RealZone(zone['vnum'])
            real_zone.age = zone['age']

            for room in zone.get('rooms') or []:
                self.restoreRoom(self.RealRoom(room['vnum']), room)

        # Now fixup references.
        player_ref = self.restorePlayers(data.get('players'))

        for (nr, (mobile, data)) in self.mobile_ref.items():
            for attribute in ('opponent', 'master', 'holding', 'mount'):
                try: o = data[attribute]
                except KeyError: pass
                else:
                    if o >= 0: o = self.mobile_ref[o][0]
                    else: o = player_ref[o]

                    try: setattr(mobile, attribute, o)
                    except AttributeError:
                        # Note: the AttributeError is for non-writable binding members (or values).
                        self.syslog('COPYOVER: Non-writable Mobile Attribute: %s' % attribute)

            try: o = data['tracking']
            except KeyError: pass
            else: mobile.tracking = o

            try: inventory = data['inventory']
            except KeyError: pass
            else:
                # todo: reverse order??
                for item in inventory:
                    self.restoreItem(mobile, item)

            # todo: equipment

        all_item_ref = {}
        while self.item_ref:
            item_ref = self.item_ref
            self.item_ref = {}

            # Save master copy.
            all_item_ref.update(item_ref)

            for (nr, (item, data)) in item_ref.items():
                try: contents = data['contents']
                except KeyError: pass
                else:
                    # todo: reverse order??
                    for contained in contents:
                        self.restoreItem(item, contained)

        self.item_ref = all_item_ref

# Programming Interface.
def getCmdlnParser():
    from optparse import OptionParser
    parser = OptionParser(prog = 'copyover')
    parser.add_option('--world')
    parser.add_option('--complete', action = 'store_true')

    parser.add_option('--format')
    parser.add_option('--analyze', action = 'store_true')
    return parser

def getValidCmdlnOptions(args):
    from stuphos.etc import parseOptionsOverSystem
    parser = getCmdlnParser()
    (options, args) = parseOptionsOverSystem(parser, args)

    if args:
        # This form of invocation doesn't accept non-option arguments,
        # so print usage and do a conventional parser-abort.
        parser.print_help()
        raise SystemExit
    else:
        return options

def parseCmdln(args):
    from stuphos.etc import parseOptionsOverSystem
    return parseOptionsOverSystem(getCmdlnParser(), args)

def getFormatHandler(filename, format = None):
    if format:
        try: return FORMATS[str(format).lower()]
        except KeyError: pass

    from os.path import splitext
    try: return FORMATS[splitext(filename)[1].lower()]
    except KeyError: pass

    return FORMATS['default']

class BinaryHandler:
    import marshal
    def loadFile(self, filename):
        return self.marshal.loads(open(filename).read().decode('base64').decode('zlib'))
    def dumpFile(self, data, filename):
        open(filename, 'wt').write(self.marshal.dumps(data).encode('zlib').encode('base64'))

class PickledHandler:
    # XXX integers automatically converted to longs??
    import pickle
    def loadFile(self, filename):
        return self.cPickle.load(open(filename))
    def dumpFile(self, data, filename):
        self.cPickle.dump(data, open(filename, 'wt'))

class JsonHandler:
    def loadFile(self, filename):
        return fromJsonFile(open(filename))
    def dumpFile(self, data, filename):
        toJsonFile(data, open(filename, 'wt'))

class PythonHandler:
    import pprint
    def loadFile(self, filename):
        return eval(open(filename).read().strip().replace('\n', ''))
    def dumpFile(self, data, filename):
        self.pprint.pprint(data, open(filename, 'wt'))

FORMATS = dict(binary = BinaryHandler(),
               pickled = PickledHandler(),
               json = JsonHandler(),
               default = PythonHandler())

FORMATS[ '.bin'] = FORMATS[ 'binary']
FORMATS['.json'] = FORMATS[   'json']
FORMATS[ '.pkl'] = FORMATS['pickled']
FORMATS[  '.py'] = FORMATS['default']

def dumpFile(data, filename, format = None):
    return getFormatHandler(filename, format = format).dumpFile(data, filename)
def loadFile(filename, format = None):
    return getFormatHandler(filename, format = format).loadFile(filename)

def dumpWorldData(options):
    return WorldFreezer(options).dump()
def loadWorldData(data):
    return WorldUnfreezer().load(data)

# Testing.
def analyzeFreezeData(data):
    from world import mobile as MobilePrototype, room as RealRoom
    memo = {}
    for zone in data.get('zones') or []:
        for room in zone.get('rooms') or []:
            room_vnum = room['vnum']
            for mobile in room.get('mobiles') or []:
                vnum = mobile['vnum']
                memo.setdefault((vnum, room_vnum), []).append(mobile)

    for ((vnum, room), instances) in memo.items():
        yield '%s: %s' % (RealRoom(room), MobilePrototype(int(vnum)))
        for data in instances:
            yield '    <%s>' % data['instance.id']
            for (name, value) in data.items():
                if name not in ['vnum', 'instance.id']:
                    yield '      %s: %r' % (name, value)

def doFreezeWorld(peer, cmd, args):
    if peer.avatar and peer.avatar.level >= 115:
        import sys
        if not hasattr(sys, 'argv'):
            sys.argv = ['']

        try: (options, args) = parseCmdln(args)
        except SystemExit:
            return True

        filename = args and args[0] or 'data.py'
        format = getFormatHandler(filename, options.format)

        if options.analyze:
            data = format.loadFile(filename)
            peer.page_string('\r\n'.join(analyzeFreezeData(data)) + '\r\n')
        else:
            data = dumpWorldData(options)
            format.dumpFile(data, filename)

        return True

try: from stuphmud.server.player import ACMD
except ImportError: pass
else: ACMD('freeze-w*orld')(doFreezeWorld)



# This is actually specific to the embedded provision.
# --
# Contains objects used across the application:
#     .network.server.zones
#     .network.server.player.shell
#

from stuphmud.server.player.interfaces.dictcmd import CmdDict

from stuphos.kernel import Girl, Script, GirlSystemModule
from stuphos.system.api import mudlog
from stuphos.runtime.architecture.api import writeprotected, wrapper
from stuphos.etc.tools import isYesValue
from stuphos.etc.tools.logs import tracebackString

#girlSystemModule = runtime.Girl.System.Module

# todo: This module contains the base object EntityWrapper but it should probably
# go into runtime.architecture.
class EntityAdapter(wrapper, writeprotected):
    # Wraps layer zero.
    _readonly_properties = []

    def __getattr__(self, name):
        try: return object.__getattribute__(self, name)
        except AttributeError as e:
            if name in self._readonly_properties:
                return self._getWrappedAttribute(name)

            raise e

    def _getWrappedAttribute(self, name):
        return getattr(self._object, name)

    def __repr__(self):
        # This must be defined manually because it's already defined on EntityWrapper :)
        return repr(self._object)

class TriggerAdapter(EntityAdapter):
    pass


# Implementation of VM.Task.
class TriggerScript(Script):
    def __init__(self, *args, **kwd):
        tracing = kwd.pop('tracing', False)
        Script.__init__(self, *args, **kwd)
        self.uncaughtError = self.handleError
        if tracing == 'profile':
            self.tracing = self.profile
        elif tracing:
            # Use operator/programmer for output.
            self.tracing = self.auditInstruction # traceToLog

    def handleError(self, task, frame, exc, traceback):
        # name = "?" if frame is None else getattr(frame.procedure, "name", "?")
        # logOperation(f'[trigger$error] {name}:' + \
        #              f' {exc[0].__name__}: {exc[1]}')


        if not self.handleOperatorError(task, frame, exc, traceback):
            (exc_class, exc_value, exc_tb) = exc
    
            # XXX :skip: mudlog doesn't log to file (console stderr)
            def mudlog(s):
                logOperation('[trigger$error] ' + s)

            mudlog('(%s: %s) %s: %s\n%s' % \
                (task.name, '?' if frame is None else \
                 getattr(frame.procedure, 'name', '?'),
                 exc_class.__name__, exc_value,
                 tracebackString(exc_tb, indent = 4)))

        raise self.Done

    def frameOne(self, *args, **kwd):
        try: return Script.frameOne(self, *args, **kwd)
        except Script.Done as e:
            # Todo: task.onComplete/frame.onComplete
            # (Right now, we're just popping the stack, which doesn't need to be done)
            try: result = self.stack.pop()[0]
            except IndexError: pass
            else:
                if result is not None:
                    pass # todo: evaluate with result

            raise e

    def traceToLog(self, frame, pos, instr, args):
        # todo: logstream to database
        mudlog('%04d %s(%s)' % (pos, getattr(instr, '__name__', '?'),
                                ', '.join(map(str, args))))

    # Invocation Methods:
    @classmethod
    def invokeTrigger(self, program, context, tracing = False,
                      programmer = None, name = None):

        # from stuphos.kernel import startTask # XXX :skip: This is somehow named wrongly in the interior import.  Where is global startTask??
        from stuphos.kernel import executeGirl
        # context.setdefault('system', GirlSystemModule.Get())

        program = program.replace('\r', '')
        program = Girl(Girl.Module, program)

        if programmer is None:
            user = None
        else:
            from stuphos.kernel import findUserByName
            user = findUserByName(programmer.principal)

        task = self.Load(tracing = tracing, environ = context, user = user) # , programmer = programmer)
        task.name = name
        task.operator = None if programmer is None else programmer.principal
        program.setEnvironment(task.environ)
        program.name = name
        # programmer.invoke(task, program)
        # startTask(task, program, programmer = programmer)

        task.addFrameCall(program, programmer = programmer)
        executeGirl(task)
        # return task

    @classmethod
    def invokeVerbCommand(self, player, target, object):
        pass

class MobileScript(TriggerScript):
    pass

class TriggerTemplate:
    # Adapter (wrapper).
    typeName = 'Trigger'
    typeCode = '--'

    class Argument:
        def __init__(self, number, name, value):
            self.number = number
            self.name = name
            self.value = value

        def setValue(self, trigger, value):
            pass

    def __init__(self, trigger):
        self.trigger = trigger

    def arguments(self):
        return iter(())
    def argumentsMapping(self):
        r = dict((arg.number, arg) for arg in self.arguments())
        return r

    def setCode(self, code):
        self.trigger.program = code


# Triggers.
class TriggerBase:
    def actTriggered(self, string):
        return False
    def departureTriggered(self):
        return False
    def arrivalTriggered(self):
        return False
    def deathTriggered(self):
        return False
    def mobileCreated(self):
        return False
    def portalOutTriggered(self):
        return False
    def portalInTriggered(self):
        return False

class MobileTrigger(TriggerBase):
    def __init__(self, program, programmer = None, tracing = None):
        self.program = program
        self.programmer = programmer
        self.tracing = tracing
    def invoke(self, *args):
        raise NotImplementedError

    FLAG_AUDIT = (1 << 0)

    @property
    def flags(self):
        yield dict(name = 'Audit', set = bool(self.tracing))

    @property
    def bitvector(self):
        return self.FLAG_AUDIT if self.tracing else 0

def prototypeTriggering(ch, callback):
    # Try triggers for this mobile instance.
    try: triggers = ch.triggers
    except AttributeError: pass
    else:
        for (i, trig) in enumerate(triggers):
            if callback(ch, trig, i):
                return True

    # Try triggers registered for the prototype.
    # Note: check identity of triggers object because
    # entity_instance (mobile_instance) overrides
    # __getattr__.
    try: ptriggers = ch.prototype.triggers
    except AttributeError: pass
    else:
        if ptriggers is not triggers:
            for (i, trig) in enumerate(triggers):
                if callback(ch, trig, i):
                    return True

def getTriggerName(object, nr, trig):
    return '%s:%d:%s' % (object, nr, getMobileTriggerType(trig).typeCode)

from random import randint
class RandomizedMobileTrigger(MobileTrigger):
    def __init__(self, chance, program):
        MobileTrigger.__init__(self, program)
        self.chance = int(chance)

    @property
    def triggerChance(self):
        return randint(0, 100) < self.chance

def getCallString(first, *args):
    return '%s(%s)' % (first, ', '.join(map(repr, args)))

class ActMobileTrigger(MobileTrigger):
    # ActMobileTrigger('*', "...")
    def __init__(self, matching, program):
        MobileTrigger.__init__(self, program)
        self.matching = matching
    def actTriggered(self, string):
        if self.matching == '*':
            return True

        import re
        return re.match(self.matching, string) is not None
        return self.matching in string

    def invoke(self, string, target, actor, item,
               indirect_actor, indirect_item,
               name):
        return MobileScript.invokeTrigger \
               (self.program,
                dict(string = string,
                     target = target,
                     actor = actor, # mud.zones.World.ZoneAccess.MobileInstance(actor),
                     item = item,
                     indirectActor = indirect_actor,
                     indirectItem = indirect_item),
                programmer = self.programmer,
                tracing = self.tracing,
                name = name)

    @classmethod
    def Event(self, original, string, target, actor, item, indirect_actor, indirect_item):
        ##    mudlog(getCallString('act', original, string, target, actor,
        ##                         item, indirect_actor, indirect_item))

        # For each call to act, as sent to a player (one with a descriptor),
        # loop over each mobile in the same room as the player, and search
        # for an act trigger.

        # Unfortunately, this doesn't do what you might expect it to: make
        # a mobile react to an act string sent to it.  This is because, most
        # of the time, no act strings are sent to mobiles because they don't
        # have playing descriptors.

        # So, this has mobiles reacting to all act strings sent to players,
        # even if it doesn't make sense that the mobile wouldn't be receiving
        # any of those strings.  Of course, it also means mobiles reacting
        # to multiple perform_act invocations per act call.

        # This is to mimic DG scripts as closely as possible (although admittedly,
        # not at all).

        def action(object, trig, nr):
            if trig.actTriggered(string):
                trig.invoke(string, target, actor, item,
                            indirect_actor, indirect_item,
                            getTriggerName(object, nr, trig))

                return True

        for ch in target.room.people:
            if ch is target:
                continue
            if ch.isPlayer:
                continue

            if prototypeTriggering(ch, action):
                return

performAct = ActMobileTrigger.Event

class MovementTrigger(RandomizedMobileTrigger):
    def invoke(self, object, mobile, origin, destination,
               direction, verb, movementNeeded, name):
        return MobileScript.invokeTrigger \
               (self.program,
                dict(object = object,
                     mobile = mobile,
                     origin = origin,
                     destination = destination,
                     direction = direction,
                     verb = verb,
                     movementNeeded = movementNeeded,
                     points = movementNeeded),
                programmer = self.programmer,
                tracing = self.tracing,
                name = name)

    def invokePortal(self, object, mobile, origin, destination, portal):
        return MobileScript.invokeTrigger \
               (self.program,
                dict(object = object,
                     mobile = mobile,
                     origin = origin,
                     destination = destination,
                     portal = portal),
                programmer = self.programmer,
                tracing = self.tracing)

class DepartureMobileTrigger(MovementTrigger):
    def departureTriggered(self):
        return self.triggerChance

    @classmethod
    def Event(self, mobile, origin, destination, direction, verb, movementNeeded):
        # mobile is what's moving -- including players, but the trigger is found on neighbors.
        # todo: invoke with triggered mobile instance handle parameter.
        def action(object, trig, nr):
            if trig.departureTriggered():
                trig.invoke(object, mobile, origin, destination,
                            direction, verb, movementNeeded,
                            getTriggerName(object, nr, trig))
                return True

        # Trigger for all NPCs that aren't currently doing the roaming.
        for ch in origin.people:
            if ch is mobile:
                continue
            if ch.isPlayer:
                continue

            if prototypeTriggering(ch, action):
                return

class PortalOutTrigger(MovementTrigger):
    def portalOutTriggered(self):
        return self.triggerChance

    @classmethod
    def Event(self, mobile, origin, destination, portal):
        def action(object, trig, nr):
            if trig.portalOutTriggered():
                trig.invokePortal(object, mobile, origin, destination, portal)
                return True

departure = DepartureMobileTrigger.Event
portalOut = PortalOutTrigger.Event

class ArrivalMobileTrigger(MovementTrigger):
    def arrivalTriggered(self):
        return self.triggerChance

    @classmethod
    def Event(self, mobile, origin, destination, direction, verb, movementNeeded):
        def action(object, trig, nr):
            if trig.arrivalTriggered():
                # import game
                # game.syslog('arrival: %r [%r]' % (object, mobile))

                trig.invoke(object, mobile, origin, destination,
                            direction, verb, movementNeeded,
                            getTriggerName(object, nr, trig))
                return True

        for ch in destination.people:
            if ch is mobile:
                continue
            if ch.isPlayer:
                continue

            if prototypeTriggering(ch, action):
                return

class PortalInTrigger(MovementTrigger):
    def portalInTriggered(self):
        return self.triggerChance

    @classmethod
    def Event(self, mobile, origin, destination, portal):
        def action(object, trig, nr):
            if trig.portalInTriggered():
                trig.invokePortal(object, mobile, origin, destination, portal)
                return True

arrival = ArrivalMobileTrigger.Event
portalIn = PortalInTrigger.Event

class DeathMobileTrigger(MobileTrigger):
    def deathTriggered(self):
        return True

    def invoke(self, victim, killer):
        return MobileScript.invokeTrigger \
               (self.program,
                dict(victim = victim,
                     killer = killer),
                programmer = self.programmer,
                tracing = self.tracing)

    @classmethod
    def Event(self, victim, killer):
        if victim.npc:
            def action(object, trig, nr):
                if trig.deathTriggered():
                    trig.invoke(victim, killer)
                    return True

            prototypeTriggering(victim, action)

slayMobile = DeathMobileTrigger.Event

class CreateMobileTrigger(MobileTrigger):
    def mobileCreated(self):
        return True

    def invoke(self, mobile, source, cause, name):
        # print(f'[create.mobile$trigger] {self.programmer}')

        return MobileScript.invokeTrigger \
               (self.program,
                dict(mobile = mobile,
                     source = source,
                     cause = cause),
                programmer = self.programmer,
                tracing = self.tracing,
                name = name)

    @classmethod
    def Event(self, mobile, source, cause):
        def action(object, trig, nr):
            if trig.mobileCreated():
                # logOperation(f'[create.mobile$trigger] #{nr}: {mobile} -- {trig.program}')
                # debugOn()

                trig.invoke(mobile, source, cause,
                            getTriggerName(object, nr, trig))

                return False # Allow pipelining of triggers.

        prototypeTriggering(mobile, action)

createMobileInstance = CreateMobileTrigger.Event


# Alteration Interface.
# This is here because it is relied on by both mud.zones and web.stuph.embedded.olc.triggers
class ActMobileTriggerTemplate(TriggerTemplate):
    typeName = 'Perform Act'
    typeCode = 'perform-act'

    methodology = \
'''
<h2>Methodology: Perform-Act</h2>

<p>
For each call to act, as sent to a player (one with a descriptor),
loop over each mobile in the same room as the player, and search
for an act trigger.
</p>

<p>
Unfortunately, this doesn't do what you might expect it to: make
a mobile react to an act string sent to it.  This is because, most
of the time, no act strings are sent to mobiles because they don't
have playing descriptors.
</p>

<p>
So, this has mobiles reacting to all act strings sent to players,
even if it doesn't make sense that the mobile wouldn't be receiving
any of those strings.  Of course, it also means mobiles reacting
to multiple perform_act invocations per act call.
</p>

<p>
This is to mimic DG scripts as closely as possible (although admittedly,
not at all).
</p>

<pre>
    def performAct(original, string, target, actor, item, indirect_actor, indirect_item):
        def action(object, trig):
            if trig.actTriggered(string):
                trig.invoke(string, target, actor, item,
                            indirect_actor, indirect_item)
                return True

        for ch in target.room.people:
            if ch is target:
                continue
            if ch.isPlayer:
                continue

            # Iterate all triggers on instance and prototype, calling action with
            # ch as object, and the individual trigger object test as trig.
            if prototypeTriggering(ch, action):
                return
</pre>
'''

    parameters = ['string', 'target', 'actor', 'item', 'indirectActor', 'indirectItem']

    class MatchingArgument(TriggerTemplate.Argument):
        number = 0
        name = 'Pattern'

        def __init__(self, trigger):
            self.value = trigger.trigger.matching
        def setValue(self, trigger, value):
            trigger.matching = value

    def arguments(self):
        yield self.MatchingArgument(self)

    @classmethod
    def CreateEmpty(self):
        return ActMobileTrigger('', '')

class ChanceTemplate(TriggerTemplate):
    class ChanceArgument(TriggerTemplate.Argument):
        number = 0
        name = 'Chance'

        def __init__(self, trigger):
            self.value = trigger.trigger.chance
        def setValue(self, trigger, value):
            trigger.chance = int(value)

    def arguments(self):
        yield self.ChanceArgument(self)

class DepartureMobileTriggerTemplate(ChanceTemplate):
    typeName = 'Departure'
    typeCode = 'departure'

    methodology = \
    '''
    Reacts probabilistically to movement of other NPCs, departing this room.
    '''

    parameters = ['object', 'mobile', 'origin', 'destination',
                  'direction', 'verb', 'points']

    @classmethod
    def CreateEmpty(self):
        return DepartureMobileTrigger(0, '')

class ArrivalMobileTriggerTemplate(ChanceTemplate):
    typeName = 'Arrival'
    typeCode = 'arrival'

    methodology = \
    '''
    Reacts probabilistically to movement of other NPCs, entering this room.
    '''

    parameters = ['object', 'mobile', 'origin', 'destination',
                  'direction', 'verb', 'points']

    @classmethod
    def CreateEmpty(self):
        return ArrivalMobileTrigger(0, '')

class PortalOutTriggerTemplate(ChanceTemplate):
    typeName = 'Portal Out'
    typeCode = 'portal-out'

    @classmethod
    def CreateEmpty(self):
        return PortalOutTrigger(0, '')

class PortalInTriggerTemplate(ChanceTemplate):
    typeName = 'Portal In'
    typeCode = 'portal-in'

    @classmethod
    def CreateEmpty(self):
        return PortalInTrigger(0, '')

class DeathMobileTriggerTemplate(TriggerTemplate):
    typeName = 'Slay Mobile'
    typeCode = 'death'

    methodology = \
    '''
    Reacts to killing of one NPC by another mobile.
    '''

    parameters = ['killer', 'victim']

    @classmethod
    def CreateEmpty(self):
        return DeathMobileTrigger('')

class CreateMobileTriggerTemplate(TriggerTemplate):
    typeName = 'Create Mobile'
    typeCode = 'create'

    parameters = ['mobile', 'source', 'cause']

    @classmethod
    def CreateEmpty(self):
        return CreateMobileTrigger('')

MOBILE_TRIGGER_TEMPLATES = \
    dict(ActMobileTrigger = ActMobileTriggerTemplate,
         DepartureMobileTrigger = DepartureMobileTriggerTemplate,
         ArrivalMobileTrigger = ArrivalMobileTriggerTemplate,
         DeathMobileTrigger = DeathMobileTriggerTemplate,
         CreateMobileTrigger = CreateMobileTriggerTemplate,
         PortalOutTrigger = PortalOutTriggerTemplate,
         PortalInTrigger = PortalInTriggerTemplate)


# For create-new-trigger select.  Ordered for recognizability.
MOBILE_TRIGGER_TYPES = [('create', CreateMobileTriggerTemplate),
                        ('death', DeathMobileTriggerTemplate),
                        ('arrival', ArrivalMobileTriggerTemplate),
                        ('departure', DepartureMobileTriggerTemplate),
                        ('portal-in', PortalInTriggerTemplate),
                        ('portal-out', PortalOutTriggerTemplate),
                        ('perform-act', ActMobileTriggerTemplate)]

def getMobileTriggerType(trigger):
    triggerType = MOBILE_TRIGGER_TEMPLATES[trigger.__class__.__name__]
    triggerType = triggerType(trigger)
    return triggerType

def getMobileTriggerTemplate(byName):
    return dict(MOBILE_TRIGGER_TYPES)[byName]

def getMobileTriggerSheet(mobile):
    try: triggers = mobile.triggers
    except AttributeError:
        triggers = []

    return MobileTriggerSheet(triggers)

class MobileTriggerSheet(dict):
    ORDER = [('create', 'create', 'Create Mobile -- Run mobile lifecycle'),
             ('death', 'death', 'Slay Mobile -- Killed by opponent'),
             ('arrival', 'arrival', 'Arriving in Room'),
             ('departure', 'departure', 'Departing from Room'),
             ('portal-in', 'portalIn', 'Portal Exit Point'),
             ('portal-out', 'portalOut', 'Portal Entrance Point'),
             ('perform-act', 'performAct', 'Act() Messaging')]

    def __init__(self, triggers):
        for (type, name, title) in self.ORDER:
            self.install(triggers, type, name)

    def install(self, triggers, type, name = None):
        if name is None:
            name = type

        for t in self.search(triggers, type):
            self.setdefault(name, []).append(t)
            break
        else:
            t = getMobileTriggerTemplate(type)
            self[name] = [t.CreateEmpty()]

    def search(self, triggers, type):
        for t in triggers:
            if getMobileTriggerType(t).typeCode == type:
                yield t

    @property
    def iterator(self):
        for (type, name, title) in self.ORDER:
            yield dict(list = self[name],
                       title = title)


# Entity Verb Commands.
class VerbGroup(CmdDict):
    def match(self, command):
        try: reg = self[command]
        except KeyError:
            return

        return reg

def verbCommand(player, command, argstr):
    ch = player.avatar
    if ch is not None:
        # if hasattr(ch.room, 'verbs'):
        #     verb = ch.room.verbs.match(command)
        #     if verb is not None:
        #         program = "call(%r, %r, %r)" % (verb.method, command, argstr)


        # A verb is a command-based action in the interactive world.
        # Todo: how to integrate with evaluateCommand?
        if isYesValue(configuration.Interpreter.evaluate_verbs):
            from stuphmud.server.player.interfaces.code.girl import PlayerScript
            return PlayerScript.evaluateVerb(player, ch, command, argstr)


        # try: spatial = ch.spatial
        # except AttributeError: pass
        # else:
        #     spatial.interpretCommandArgs(player, command, argstr)




from . import CommandMessage, RegisterCommandClass

class ShellCommand:
    def parseCommandLine(self, cmdln):
        # XXX horrible parsing
        return cmdln.split()

class AdminScriptCommandMessage(CommandMessage, ShellCommand):
    def Execute(self):
        argv = self.parseCommandLine(self.payload)

        # Try to find argv[0] in etc/admin_scripts directory,
        # execute as a shell command with arguments.

class SystemShellCommandMessage(CommandMessage, ShellCommand):
    def Execute(self):
        from stuphos.runtime.registry import getObject
        argv = self.parseCommandLine(self.payload)
        shell = getObject('System::Shell::API')
        if shell:
            from stuphos import enqueueHeartbeatTask
            from stuphos.system.api import syslog

            pipe = shell.doCommandHeadless(argv[0], *argv[1:])
            def processOutput():
                result = pipe.read()
                enqueueHeartbeatTask(syslog, 'System Shell Result: %s' % result)

            enqueueHeartbeatTask(processOutput)

RegisterCommandClass('application/x-stuph-system-shell-command', SystemShellCommandMessage)
RegisterCommandClass('application/x-stuph-admin-script', AdminScriptCommandMessage)



#!/usr/bin/python2.4
 #-
 # Copyright (c) 2008 Clint Banis (hereby known as "The Author")
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions
 # are met:
 # 1. Redistributions of source code must retain the above copyright
 #    notice, this list of conditions and the following disclaimer.
 # 2. Redistributions in binary form must reproduce the above copyright
 #    notice, this list of conditions and the following disclaimer in the
 #    documentation and/or other materials provided with the distribution.
 # 3. All advertising materials mentioning features or use of this software
 #    must display the following acknowledgement:
 #        This product includes software developed by The Author, Clint Banis.
 # 4. The name of The Author may not be used to endorse or promote products
 #    derived from this software without specific prior written permission.
 #
 # THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS
 # ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 # TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 # PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS
 # BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 # CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 # SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 # INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 # CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 # ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 #

'Simulate a MUDOperations struct and register it with the game module via ctypes.'

from ctypes import c_void_p, c_int, py_object
from ctypes import CFUNCTYPE, PYFUNCTYPE, POINTER
from ctypes import pythonapi, byref, cast, addressof

## Business Logic.
from .structs import EventOperations, MUDOperations, AUTOFUNCTYPE
linkModuleFunction = AUTOFUNCTYPE(EventOperations.P, MUDOperations.P)

PyCObject_AsVoidPtr = pythonapi.PyCObject_AsVoidPtr
PyCObject_AsVoidPtr.restype = c_void_p
PyCObject_AsVoidPtr.argtypes = [py_object]

def registerGameModule(ops):
        '''
        Imports game module and calls its registration method with the supplied ops,
        returning an event bridge.
        '''

        try: from stuphos.system.api import game
        except ImportError:
                register = synthesizeGameModule().register
        else:
                register = game.register

        link = linkModuleFunction(PyCObject_AsVoidPtr(register))
        return link(byref(ops))


## Application Ware.
class API:
        def __init__(self, OPSClass):
                self.ops = OPSClass(*self.getBindings(OPSClass._fields_))

        def getBindings(self, fields):
            for info in fields:
                binding = self.Binding(info, self)
                # yield info[1](binding)

                # Debugging
                try: yield info[1](binding)
                except Exception as e:
                    from traceback import print_exc as traceback
                    from pdb import set_trace as enterDebugger

                    traceback()
                    enterDebugger()

                    raise e

        class Binding:
                # Store field info bound to the api.
                def __init__(self, field_info, api):
                        self.procname, self.procdef = field_info
                        self.api = api

                def __call__(self, *args, **kwd):
                    from ph import log as mudlog
                    mudlog(self.api.callString(self.procname, args, kwd))

                    # Construct new default return type, as expected.
                    return addressof(self._default_result_)

                @property
                def _default_result_(self):
                    try: result = self.__result
                    except AttributeError:
                        try: result = self.procdef._data_type_()
                        except AttributeError:
                            result = self.procdef._restype_()

                        self.__result = result

                    return result

        def callString(self, name, args, kwd):
            scope = self.__class__.__name__
            if name.startswith('getvalue_'):
                return '%s::%s' % (scope, name[9:])

            if name.startswith('setvalue_'):
                return '%s::%s = %s' % (scope, name[9:], args[0])

            args = list(map(repr, args))
            args.extend('%s = %r' % i for i in kwd.items())
            args = ', '.join(args)

            return '%s::%s(%s)' % (scope, name, args)

        def __getattr__(self, name):
            namestr = str(name)
            try: getvalue = getattr(self.__dict__['ops'],
                                    'getvalue_' + namestr)

            except (AttributeError, KeyError):
                try: return self.__dict__[name]
                except KeyError:
                    # Re-raise so that architecture works.
                    raise AttributeError(name)

            return getvalue()

        def __setattr__(self, name, value):
            namestr = str(name)
            try: setvalue = getattr(self.__dict__['ops'],
                                    'setvalue_' + namestr)

            except (AttributeError, KeyError):
                self.__dict__[name] = value
                return

            setvalue(value)

## Compatability.
def synthesizeGameModule():
        # Cell for the register callback.
        try: from stuphos.system.api import game
        except ImportError:
                import new, sys
                game = sys.modules['game'] = new.module('game')

        # Build registration object.
        PyCObject_FromVoidPtr = pythonapi.PyCObject_FromVoidPtr
        PyCObject_FromVoidPtr.restype = py_object
        PyCObject_FromVoidPtr.argtypes = [c_void_p]

        def registerCallback(mudApi):
                from ph import log as mudlog
                mudlog('[Registering Game Module (%r)]' % mudApi)
                game.mudApi = mudApi

                # Remember reference.
                game.eventBridge = EventOperations()
                return addressof(game.eventBridge)

        # Remember reference.
        game.link = linkModuleFunction(registerCallback)
        game.register = PyCObject_FromVoidPtr(game.link)

        return game

## Test Facility.
class Main(API):
        def __init__(self):
                API.__init__(self, MUDOperations)
                self.eventBridge = registerGameModule(self.ops)

        def __repr__(self):
                return '%s\n  ops: %r\n  eventBridge: %r' % \
                       (self.__class__.__name__,
                        self.ops, self.eventBridge)

        __str__ = __repr__

def main():
    synthesizeGameModule()

    if 0:
        global ops
        ops = Main()
        print(ops)

if __name__ == '__main__':
    main()



# Pentacle - The out-of-band binary telnet protocol interface.
# Note: this should probably go in mud.management, because it qualifies as a high-
# level service, but it's also trying to be low-level.
# Todo: move into mud.management.
from stuphos.runtime.facilities import Facility

# Protocol:
#   Challenge/Handshake for hashing password but only sending the hash for auth.

class Pentacle(Facility):
    NAME = 'Network::Telnet::Pentacle'

    @classmethod
    def create(self):
        return newComponent(self)

    # Component Architecture.
    def onTelnetCommand(self, ctlr, peer, tc):
        pass



# Tool for uploading/modifying player files via a telnet connection.
# Copyright 2011 Clint Banis.  All rights reserved.
from time import sleep
from configparser import DEFAULTSECT, ConfigParser
from os.path import expanduser
from telnetlib import Telnet
from pdb import set_trace, runcall
from optparse import OptionParser
import sys
import traceback

CONFIG_NAMES = ('username', 'password', 'host', 'port', 'operation',
                'remote-filename', 'local-filename', 'new-filename',
                'message-format', 'delay', 'show-output', 'prefix',
                'key-file')
CONFIG_DEFAULTS = dict((n, '') for n in CONFIG_NAMES)
CONFIG_FILE = '~/.telnetftp_client.cfg'

DEFAULT_TELNET_MODULE = 'mud.player.telnet'

def getCmdlnParser():
    parser = OptionParser()
    parser.add_option('--username')
    parser.add_option('--password')
    parser.add_option('--host')
    parser.add_option('--port')
    parser.add_option('--operation')
    parser.add_option('--remote-filename')
    parser.add_option('--command', dest = 'remote_filename')
    parser.add_option('--local-filename')
    parser.add_option('--new-filename')
    parser.add_option('--key-file', '--keyfile')
    parser.add_option('--telnet-module', default = DEFAULT_TELNET_MODULE)
    parser.add_option('--message-format', default = 'json')
    parser.add_option('-g', '--debug', action = 'store_true')
    parser.add_option('-w', '--delay', type = int, default = 0)
    parser.add_option('-C', '--config', default = CONFIG_FILE)
    parser.add_option('--section')
    parser.add_option('--show-output')
    parser.add_option('--prefix')
    return parser

def parse_cmdln(argv = None, parser = None):
    if parser is None:
        global __parser
        try: parser = __parser
        except NameError:
            parser = __parser = getCmdlnParser()

    return parser.parse_args(argv)

# XXX: This loads file config over cmdline options, instead of the other way around.
def loadConfig(options, config_file, section):
    if options.debug:
        set_trace()

    cfg = ConfigParser(defaults = CONFIG_DEFAULTS)
    cfg.read([config_file])

    for o in CONFIG_NAMES:
        value = cfg.get(section, o)
        if value is not '':
            o = o.replace('-', '_')
            setattr(options, o, value)

    # Validate.
    options.delay = int(options.delay)
    return options

def getConfigFile(config_file):
    i = config_file.rfind(':')
    if i >= 0:
        section = config_file[i+1:]
        config_file = config_file[:i]
    else:
        section = DEFAULTSECT

    return (config_file, section)

# Encryption Routines.
def LoadKeyfile(keyfile):
    return open(keyfile).read().strip()

def getCipher(key):
    from crypto.cipher.trolldoll import Trolldoll
    alg = Trolldoll(ivSize = 160)
    alg.setPassphrase(key) # .setKey(key)
    return alg

def encrypt(key, content):
    return getCipher(key).encrypt(content)
def decrypt(key, content):
    return getCipher(key).decrypt(content)

def encryptArmored(key, content):
    return encrypt(key, content).encode('base64').replace('\n', '')
def decryptArmored(key, content):
    return decrypt(key, content.decode('base64'))

def GetMessageDigest(key_iv, player_name_salt, payload):
    import hmac
    hash = hmac.new(key_iv)
    hash.update(player_name_salt)
    hash.update(payload)
    return hash.hexdigest()

# Session.
def login(options):
    telnet = Telnet(options.host, options.port)
    telnet.write('%s\r\n' % options.username)
    telnet.write('%s\r\n' % options.password)
    if options.delay:
        sleep(options.delay)

    telnet.write('telnet-control manage\r\n')
    return telnet

def logout(telnet):
    telnet.write('quit\r\n0\r\n')

def get_send_sudata_routine(module_name = None):
    module = __import__(module_name or DEFAULT_TELNET_MODULE, globals(), locals(), [''])
    return module.send_sudata

def sendData(conn, data, send_sudata_proc, delay = False):
    send_sudata_proc(conn.sock, data)

    if delay:
        sleep(delay)

def printOutput(output, stream):
    output = output.split('\n')
    output = ' %  ' + '\n %  '.join(output)
    print(output, file=stream)

def getPrefixedOperation(options):
    if options.prefix:
        return '%s:%s' % (options.prefix, options.operation)
    return options.operation

def buildCmdlnOperation(options, args):
    if options.config:
        (config_file, section) = getConfigFile(options.config)
        loadConfig(options, expanduser(options.config), section)

    if options.operation not in ['list']:
        assert options.remote_filename

    assert options.host
    assert options.port
    assert options.username
    assert options.password

    keyfile = options.key_file
    if keyfile:
        key = LoadKeyfile(keyfile)

    # Validate operation.
    if options.operation in ['create', 'append']:
        filename = options.local_filename
        assert filename
        content = open(filename).read()

        if keyfile:
            assert options.operation == 'create'
            content = encryptArmored(key, content)

    elif options.operation == 'rename':
        newname = options.new_filename
        assert newname

    elif options.operation not in ['delete', 'touch', 'truncate', 'list', 'stat', 'view',
                                   'decrypt', 'execute', 'executefile']:

        raise NameError(options.operation)

    if options.operation in ['execute', 'executefile']:
        assert key

    # Validate format.
    if options.message_format == 'json':
        from simplejson import dumps
    elif options.message_format == 'pickle':
        from pickle import dumps
    else:
        raise NameError(options.message_format)

    # Configure Rest.
    if options.debug:
        set_trace()

    if options.show_output == '--':
        stream = sys.stdout
    elif options.show_output:
        stream = open(options.show_output, 'w')

    # Build Operation.
    command = [getPrefixedOperation(options)]

    if options.operation in ['execute', 'executefile']:
        command.append(GetMessageDigest(key, options.username, options.remote_filename))

    if options.operation != 'list' or options.remote_filename:
        command.append(options.remote_filename)

    if options.operation in ['create', 'append']:
        command.append(content)
    elif options.operation == 'rename':
        command.append(newname)

    command = dumps(command)
    command = '%s:%s' % (options.message_format.upper(), command)

    return command

def process_batch_file(filename):
    this = []
    for line in open(filename):
        line = line.rstrip()
        if line.endswith('\\'):
            this.append(line[:-1].rstrip())
        else:
            yield ' '.join(this)
            this = []

    if this:
        yield ' '.join(this)

def main(argv = None):
    # Configure send proc & generate message.
    (options, args) = parse_cmdln(argv)
    commands = [buildCmdlnOperation(options, args)]

    if options.batch_file:
        for cmdln in process_batch_file(options.batch_file):
            commands.append(buildCmdlnOperation(*parse_cmdln(cmdln)))

    # General telnet session.
    send_sudata_proc = get_send_sudata_routine(options.telnet_module)
    telnet = login(options)
    try:
        # Send via Telnet.
        print('Sending %d bytes (%dk/%dM)' % (len(command),
                                                             len(command) / 1024.,
                                                             len(command) / 1024 / 1024.), file=sys.stderr)

        print(repr(command[:80]), file=sys.stderr)

        # Multiple-commands.
        for cmd in commands:
            sendData(telnet, cmd, send_sudata_proc, options.delay)

        # Output.
        if options.show_output:
            while True:
                output = telnet.read_very_eager()
                if not output:
                    break

                printOutput(output, stream)

    except: traceback.print_exc()
    finally:
        logout(telnet)
        if options.show_output:
            printOutput(telnet.read_all(), stream)

if __name__ == '__main__':
    main()



# Player-Oriented File Notebook.
# Todo: adapt this to item-oriented specproc!
#
from stuphmud.server.player import *
from stuphmud.server.player.db.managedfs import *
from stuphmud.server.player.db.managedfs import getPlayerFilename, validatePlayername
from stuphos.etc import columnize, xorString as xor
from stuphos import getConfig

from stuphos.db import dbCore
from stuphos.db.orm import PlayerNotebook

import pickle
import shelve
import string
import re

# Default Notebooks.
PLAYER_NOTEBOOK_FOLDER = 'library'
PLAYER_NOTEBOOK_NAME = 'notebook'

DEFAULT_PAGE_NAME = 'Front-Matter'
PAPER_VNUM = 3

PAGE_NAME_PATTERN = re.compile(r"^a page named: '([^']+)'$")
def parsePageName(name):
    m = PAGE_NAME_PATTERN.match(name)
    if m is not None:
        return m.group(1)

# Disk Usage.
KILOBYTE = 1024
MEGABYTE = KILOBYTE * 1024
GIGABYTE = MEGABYTE * 1024
TERABYTE = GIGABYTE * 1024
EXABYTE  = TERABYTE * 1024
PETABYTE = EXABYTE  * 1024

BYTE_LEVELS = [(PETABYTE, 'Pb'),
               (EXABYTE,  'Eb'),
               (TERABYTE, 'Tb'),
               (GIGABYTE, 'Gb'),
               (MEGABYTE, 'Mb'),
               (KILOBYTE, 'Kb')]

def getHumanBytes(bytes):
    for (level, code) in BYTE_LEVELS:
        if bytes >= level:
            return '%.2f%s' % (bytes / level, code)

    return '%d bytes' % bytes

def calculateAllFileSizes(path):
    from os import stat
    from glob import glob

    total = 0
    for fn in glob(path + '*'):
        try: total += stat(fn).st_size
        except: pass

    return total

# Some support for opening other's notebooks:
def getNotebookFilename(plrname, name = None, folder = None):
    filename = getPlayerFilename(validatePlayername(plrname),
                                 name or PLAYER_NOTEBOOK_NAME,
                                 folder = folder or PLAYER_NOTEBOOK_FOLDER)

    assert filename is not None # , UnconfiguredException
    return filename

class Notebook:
    class FileModel:
        @classmethod
        def Open(self, peer, name = None, folder = None):
            # Uses checked particular variant.
            self.fs_path = getCheckedPlayerFilename(peer, name or PLAYER_NOTEBOOK_NAME,
                                                    folder = folder or PLAYER_NOTEBOOK_FOLDER)

            return self(shelve.open(self.fs_path))

        @classmethod
        def OpenForPlayer(self, plrname, name = None, folder = None):
            return self(shelve.open(getNotebookFilename(plrname, name = name, folder = folder)))

        def __init__(self, shelf):
            self.shelf = shelf
            self.close = shelf.close

    class ORMModel:
        @classmethod
        def Open(self, peer, name = None, folder = None):
            return self(self.ORMShelf(peer.avatar.name, name or PLAYER_NOTEBOOK_NAME,
                                      folder = folder or PLAYER_NOTEBOOK_FOLDER))

        @classmethod
        def OpenForPlayer(self, plrname, name = None, folder = None):
            # ('python.db', managedfs.PROGRAMS_DIR)
            return self(self.ORMShelf(plrname, name or PLAYER_NOTEBOOK_NAME,
                                      folder = folder or PLAYER_NOTEBOOK_FOLDER))

        def __init__(self, shelf):
            self.shelf = shelf

        def close(self):
            return self.shelf.close()

        class ORMShelf:
            def __init__(self, player, recordBaseName, folder = None):
                self.player = player
                self.recordBaseName = recordBaseName
                self.folder = folder

            PAGE_PICKLED = (1 << 0)

            @property
            def dbNamespace(self):
                return getConfig('database', 'Notebook')

            def __getitem__(self, name):
                # Return string content.
                with dbCore.hubThread(self.dbNamespace):
                    try:
                        for page in PlayerNotebook.selectBy(player = self.player, folder = self.folder, name = name):
                            if page.flags & self.PAGE_PICKLED:
                                return pickle.loads(page.content)

                            return page.content

                    except sqlite.module.dberrors.OperationalError:
                        pass

                raise KeyError(name)

            def __setitem__(self, name, value):
                if not isinstance(value, str):
                    value = pickle.dumps(value)
                    flags = self.PAGE_PICKLED
                else:
                    flags = 0

                with dbCore.hubThread(self.dbNamespace):
                    # Upsert.
                    for page in PlayerNotebook.selectBy(player = self.player, folder = self.folder, name = name):
                        page.content = value
                        page.flags = flags
                        break
                    else:
                        page = PlayerNotebook(player = self.player, folder = self.folder,
                                              name = name, content = value, flags = flags)

                    page.sync()

            def __delitem__(self, name):
                with dbCore.hubThread(self.dbNamespace):
                    PlayerNotebook.deleteBy(player = self.player, folder = self.folder, name = name)

            def __contains__(self, name):
                try: self[name]
                except KeyError:
                    return False

                return True

            def iterkeys(self):
                with dbCore.hubThread(self.dbNamespace):
                    for page in PlayerNotebook.selectBy(player = self.player, folder = self.folder):
                        yield page.name

            def iteritems(self):
                with dbCore.hubThread(self.dbNamespace):
                    for page in PlayerNotebook.selectBy(player = self.player, folder = self.folder):
                        yield (page.name, page.content)

            def items(self):
                return list(self.iteritems())

            def keys(self):
                return list(self.iterkeys())

            def close(self):
                pass
            def sync(self):
                pass

    Model = ORMModel

    class PlayerInterface(Model):
        # Command routines: they close the notebook explicitly on change.
        def doShow(self, peer):
            keys = list(self.shelf.keys())
            if not keys:
                playerAlert('No pages have been entered into the notebook.')

            width = max(list(map(len, keys)))
            # width = min(20, width)

            nr_cols = (120 / width)
            nr_cols = max(6, nr_cols)
            nr_cols = min(2, nr_cols)

            string = columnize(keys, nr_cols, width + 2)
            string = string.split('\n')
            string = [_f for _f in string if _f]
            string = '    ' + '\n    '.join(string)
            string = string.replace('\n', '\r\n')
            string += '\r\n'

            try: bytes = self.getDiskSize()
            except: bytes = 0

            bytes = getHumanBytes(bytes)

            string = 'Notebook [%d pages; %s]\r\n%s' % (len(keys), bytes, string)
            peer.page_string(string)

        def doView(self, peer, name = DEFAULT_PAGE_NAME):
            try: peer.page_string(self.shelf[name] or 'No text.\r\n')
            except KeyError:
                playerAlert('&rNo such page name: &w%r&N' % name)

        def doDelete(self, peer, name = DEFAULT_PAGE_NAME):
            try: del self.shelf[name]
            except KeyError:
                playerAlert('&rNo such page name: &w%r&N' % name)
            else:
                self.close()

        def doEdit(self, peer, name = DEFAULT_PAGE_NAME):
            def enterNotebookPage(peer, text):
                self.shelf[name] = text
                self.close()
                print('&w%r&N entered into notebook.' % name, file=peer)

                # delete messenger reference?
                # return True

            try: original = self.shelf[name]
            except KeyError:
                original = ''

            peer.messenger = enterNotebookPage
            if hasattr(peer, 'richEditor'):
                peer.richEditor(original, header = name)
            else:
                peer.editString(original)

        def doCopy(self, peer, source, copy, overwrite = False):
            if copy in self.shelf:
                if overwrite not in ['overwrite']:
                    playerAlert('&rYou must specify "overwrite" afterwards to do this.&N')

                # Note: overwrite could be the name of a piece of paper in inventory, or
                # as a board message!!

            self.shelf[copy] = self.shelf[source]
            self.close()

            playerAlert('Copied page %s to %s.' % (source, copy))

        def doTearOut(self, peer, name):
            try: content = self.shelf[name]
            except KeyError:
                playerAlert('&rUnknown page: %r&N' % name)

            import world
            paper = world.item(PAPER_VNUM).load(peer.avatar)
            paper.detailed_description = content
            paper.keywords = name + ' paper page'
            paper.name = "a page named: '%s'" % name

            # peer.avatar.save()

            del self.shelf[name]
            self.close()

            playerAlert('You tore out page %r' % name)

        def doPasteIn(self, peer, keyword, force = False):
            page = peer.avatar.find(keyword, inventory = True)
            if page is None:
                playerAlert("You don't have a %r to paste in!" % keyword)

            itemName = page.name
            if page.type != 'Note':
                playerAlert('%s is not something you can paste into your notebook.' % \
                            itemName.capitalize())

            pageName = parsePageName(itemName)
            if pageName is None:
                if force is False or not peer.avatar.supreme:
                    playerAlert("You can't seem to figure out how to paste %r into your notebook." % \
                                itemName)

                pageName = force

            if pageName in self.shelf:
                playerAlert('There is already a page named %r in your notebook.' % pageName)

            self.shelf[pageName] = page.detailed_description
            self.close()

            page.extract()

            playerAlert('You paste %s into your notebook.' % itemName)

        def doEncrypt(self, peer, name, key = None):
            try: content = self.shelf[name]
            except KeyError:
                playerAlert('No page named: %r' % name)

            if key is None:
                key = '' # Todo: get personal key??  Or, key by name.

            # Encrypt/Bind.
            ciphertext = xor(content, key)
            bound = ciphertext.encode('zlib').encode('base64')

            from email.message import Message
            msg = Message()
            msg.set_type('text/stuph-binding')

            msg['Content-Transfer-Encoding'] = 'zlib; base64'
            msg.set_payload(bound)

            content = msg.as_string()

            # Put back into notebook.
            self.shelf[name] = content
            self.close()

            playerAlert('Bound %r.' % name)

        def doDecrypt(self, peer, name, key = None):
            try: content = self.shelf[name]
            except KeyError:
                playerAlert('No page named: %r' % name)

            if key is None:
                key = '' # Todo: get personal key??  Or, key by name.

            # Decrypt/Unbind.
            from email import message_from_string
            msg = message_from_string(content)

            if msg.get_content_type() != 'text/stuph-binding':
                playerAlert('The page is not bound.')

            encoding = msg.get('Content-Transfer-Encoding')
            if encoding != 'encrypted; zlib; base64':
                playerAlert('The page is not bound properly: %s.' % encoding)

            content = msg.get_payload()
            ciphertext = content.decode('base64').decode('zlib')

            content = xor(ciphertext, key)

            # Put back into notebook.
            self.shelf[name] = content
            self.close()

            playerAlert('Unbound %r.' % name)

        def doEmail(self, peer):
            # actually, just mud-mail?
            playerAlert('Not yet supported!')

        # About this store (version, format, capabilities, signatures, etc.)
        def getMeta(self):
            try: return self.shelf['__meta__']
            except KeyError:
                return {}

        def getDiskSize(self):
            # We can't know, so we try to add up all files in the directory.
            return calculateAllFileSizes(self.fs_path)

    class ProgrammaticInterface(Model):
        def pageNames(self):
            return iter(self.shelf.keys())

        def getPageContent(self, name):
            return self.shelf[name]
        def setPageContent(self, name, value):
            self.shelf[name] = value

        __iter__ = pageNames
        __getitem__ = getPageContent
        __setitem__ = setPageContent

        def __enter__(self):
            return self
        def __exit__(self, etype = None, value = None, tb = None):
            self.close()

        @classmethod
        def IntelligentOpen(self, player, name = None, folder = None):
            from world import mobile as MobileType, peer as PeerType
            if isinstance(player, MobileType):
                assert player.isPlayer
                player = player.name

            if isinstance(player, str):
                return self.OpenForPlayer(player, name = None, folder = None)

            assert isinstance(player, PeerType)
            return self.Open(player, name = None, folder = None)

    ##    @runtime.api(runtime.Player.Notebook.API)
    ##    class API:
    ##        Open = ProgrammaticInterface.IntelligentOpen

# Install Runtime API.
runtime.Player.Notebook.API.Open(lambda:Notebook        \
                                 .ProgrammaticInterface \
                                 .IntelligentOpen)

@ACMD('note*book')
def doPlayerNotebook(peer, cmd, argstr):
    args = () if not argstr else argstr.split()
    larg = '' if not args else args[0].lower()

    try:
        OpenPlayerNotebook = Notebook.PlayerInterface.Open

        if larg in ['', 'show', 'list']:
            OpenPlayerNotebook(peer).doShow(peer, *args[1:])
        elif larg in ['view']:
            OpenPlayerNotebook(peer).doView(peer, *args[1:])
        elif larg in ['delete']:
            OpenPlayerNotebook(peer).doDelete(peer, *args[1:])
        elif larg in ['edit']:
            OpenPlayerNotebook(peer).doEdit(peer, *args[1:])
        elif larg in ['encrypt', 'bind']:
            OpenPlayerNotebook(peer).doEncrypt(peer, *args[1:])
        elif larg in ['decrypt', 'unbind']:
            OpenPlayerNotebook(peer).doDecrypt(peer, *args[1:])
        elif larg in ['email']:
            OpenPlayerNotebook(peer).doEmail(peer, *args[1:])
        elif larg in ['copy']:
            OpenPlayerNotebook(peer).doCopy(peer, *args[1:])
        elif larg in ['tearout', 'tear-out', 'tear']:
            OpenPlayerNotebook(peer).doTearOut(peer, *args[1:])
        elif larg in ['pastein', 'paste-in', 'paste']:
            OpenPlayerNotebook(peer).doPasteIn(peer, *args[1:])
        else:
            print('Unknown subcommand: %r' % args[0], file=peer)

    except PlayerAlert as e:
        e.deliver(peer)

    return True

# Attempt at specproc.
class SpecialNotebook:
    def __init__(self, filename):
        self.filename = filename

    def Open(self):
        OpenSpecialNotebook = Notebook.PlayerInterface
        return OpenSpecialNotebook(self.filename)

    ACTION_MAP = dict(look = 'doShow', examine = 'doShow',
                      read = 'doView', write = 'doEdit',
                      junk = 'doDelete', mail = 'doEmail')
                      # encrypt/bind
                      # decrypt/unbind
                      # copy
                      # tearout
                      # paste

    def __call__(self, this, ch, cmd, argstr):
        peer = ch.peer
        if peer is not None:
            try: actionName = self.ACTION_MAP[cmd.name]
            except KeyError: pass
            else:
                nb = self.OpenNotebook()
                action = getattr(nb, actionName)

                # Something we recognize, now make sure we're targeting this object.
                args = () if not argstr else argstr.split()
                if len(args) and ch.find(args[0]) is this:
                    try: action(peer, *args)
                    except PlayerAlert as e:
                        e.deliver(peer)

                    return True

# Directory-based notebook access.
##    from stuphlib.directory import Directory
##    PLRFILES = Directory.FromTree('PLRFILES',
##            '''
##            (partition-map):
##                format = 'plrfiles/%(name)s/%(folder)s/%(name)s.db
##            ''')




# VarDB
# Requires WRLC.
import op.data.sqlite

from stuphos.runtime.architecture.api import writeprotected, representable
from stuphos.runtime.architecture import newClassObject
from stuphos.kernel import AutoMemoryMapping, AutoMemoryNamespace, baseStringClass, vmCurrentTask

from stuphos import getConfig
from . import orm, dbCore
from .orm import VariableDBNativeTool, createSQLObjectTable as createSqlTable

from contextlib import contextmanager
from datetime import datetime

@contextmanager
def database():
	with dbCore.hubThread(configuration.VariableDB.namespace) as o:
		yield o

def buildSqlObject(tableName, registry, cfg):
	# def populateNS(ns):
	# 	ns.update(cfg)

	try: return newClassObject(tableName, (sqlite.Object,), cfg) # exec_body = populateNS)
	except ValueError:
		# XXX The registry will need to be flushed if the table definition
		# changes in the database schema.
		from sqlobject import classregistry # sqlite.module.
		return classregistry.registry(registry).getClass(tableName)

# def createSqlTable(table):
# 	from sqlobject.dberrors import OperationalError
# 	try: table.createTable()
# 	except OperationalError:
# 		pass

# Database Flags.
DBFLAG_NOACCOUNT = (1 << 0)


# class VariableDB(writeprotected, AutoMemoryMapping, representable):
class VariableDB(writeprotected, representable):
	# _preserve_memory_binding = True
	_memoryObjectManaged = True

	DBFLAG_NOACCOUNT = DBFLAG_NOACCOUNT


	@classmethod
	def _Select(self, name):
		with database():
			yield from orm.DatabaseConfiguration.selectBy(name = name)

	@classmethod
	def _Names(self):
		with database():
			for db in orm.DatabaseConfiguration.select():
				yield db.name

	@classmethod
	def _Open(self, name, **cfg):
		for dbConf in self._Select(name):
			# Note: relies on wrlc
			handle = sqlite.pathOpen(io.here(dbConf.path))
			return self(name, dbConf.hard, dbConf.flags, handle, **cfg)

		raise NameError(name)

	@classmethod
	def _Create(self, name, path, hard, flags = 0):
		for dbConf in self._Select(name):
			raise NameError(name)

		return orm.DatabaseConfiguration(name = name, path = path,
										 hard = hard, flags = flags)

	class _Namespace(writeprotected, representable):
		def __init__(self, object):
			self._object = object

		def __getattr__(self, name):
			try: return object.__getattribute__(self, name)
			except AttributeError:
				try: return self._object[name]
				except KeyError:
					raise AttributeError(name)


	# __public_members__ = ['name', 'tables']

	def __init__(self, name, hard, flags, db, **cfg):
		# AutoMemoryMapping.__init__(self)
		self._tables = dict()
		self._name = name
		self._hard = hard
		self._flags = flags or 0
		self._db = db

		self._setAttribute('tables', self._Namespace(self))
		for (name, t) in cfg.items():
			t._set_database(self)
			self._tables[name] = t

	@property
	def name(self):
		return self._name

	@property
	def path(self):
		r = ['system:vardb', 'path'] + self._name.split('/')
		vmCurrentTask().checkAccessSystem(r, 'read')

		for dbConf in self._Select(self._name):
			return dbConf.path

		raise NameError(self._name)

	@property
	def sqliteConnectURI(self):
		path = io.path(self.path)
		return sqlite.sqliteURIFromPath(str(path if path.absolute else io.here(path)))

	@property
	def noAccount(self):
		return self._flags & DBFLAG_NOACCOUNT


	def __setitem__(self, item, value):
		raise NotImplementedError('Operation not permitted')

	def __getitem__(self, name):
		return self._tables[name]

	def __getattr__(self, name):
		try: return object.__getattribute__(self, name)
		except AttributeError as e:
			try: return self[name]
			except KeyError:
				raise e

	def _calculateSize(self):
		return sum(table.size for table in self._tables.values())

	@property
	def size(self):
		try: return self._size
		except AttributeError:
			self._size = size = self._calculateSize()
			return size

	@size.deleter
	def size(self):
		try: del self._size
		except AttributeError:
			pass

	# def _memoryChange(self, table, entity, new, old):
	# 	pass


	def _createTables(self):
		# note: this isn't used because tables are created during class initialization.
		for table in self.values():
			if isinstance(table, self.Table):
				createSqlTable(table._tableClass)
				createSqlTable(table._accountingTable)


	# todo: make private?
	class Table(writeprotected):
		# __public_members__ = ['name']

		__name__ = 'table'

		def __init__(self, name, columns):
			self._name = name
			self._definition = columns

			self._checkColumnObjectNames()

		def __repr__(self):
			return f'<table {self._name}>'

		@classmethod
		def _accounting_definition(self, sqlmeta):
			# Load sqlobject and def at runtime.
			from sqlobject import IntCol
			return dict(object = IntCol(),
		                size = IntCol(),
		                sqlmeta = sqlmeta)

		@property
		def name(self):
			return self._name

		@property
		def database(self):
			return self._database

		def _set_database(self, db):
			# print string.call('newClassObject', name, (sqlite.Object,), **cfg)
			# from sqlite.module.classregistry import registry
			# registry(db.name)

			assert isinstance(db, VariableDB)
			self._database = db

			noAcct = db.noAccount

			class sqlmeta:
				registry = db.name
				lazyUpdate = True

			cfg = dict(self._definition, sqlmeta = sqlmeta)

			# XXX When calling from the native interface, the cfg will have objects
			# of class VariableDBNativeTool._Column which would make this code wrong.
			# But when built from the structural 'table' function defined below, it
			# builds the _columnClass for the type per column name, out of the Open
			# call connecting database, which is the right level for sqlobject.
			self._tableClass = buildSqlObject(self.name, db.name, cfg)

			if not noAcct:
				self._accountingTable = buildSqlObject \
					('%s_acct' % self.name, db.name,
					 self._accounting_definition(sqlmeta))

			self._tableClass._connection = db._db
			if not noAcct:
				self._accountingTable._connection = db._db

			createSqlTable(self._tableClass)
			if not noAcct:
				createSqlTable(self._accountingTable)

		def _calculateSize(self):
			return sum(row.size for row in self._accountingTable.select())

		@property
		def size(self):
			try: return self._size
			except AttributeError:
				self._size = size = self._calculateSize()
				return size

		@size.deleter
		def size(self):
			try: del self._size
			except AttributeError:
				pass

			del self._database.size

		@property
		def _columnNames(self):
			return (n for (n, c) in self._definition)

		def _checkColumnObjectNames(self):
			if any(n.startswith('_') for n in self._columnNames):
				raise AttributeError('Column names cannot start with undescore')

		@property
		def _writables(self):
			return list(self._columnNames)

		@property
		def columns(self):
			return vmCurrentTask().sequence(self._columnNames)

		def columnType(self, name):
			pass # return self._definition[name]

		def __call__(self, *args):
			# runtime.call.System.Engine.IO(args)
			# debugOn()
			return self._Entity(self, values = self._toKeywordColumns(args))

		def _toKeywordColumns(self, values):
			from sqlobject import DateTimeCol

			row = dict()

			d = self._definition
			for i in range(len(d)):
				(n, c) = d[i]
				v = values[i]

				if isinstance(v, baseStringClass):
					v = str(v)

				# Inline internal column value data type conversion(s).
				if isinstance(c, DateTimeCol):
					if not isinstance(v, float):
						raise TypeError('Column #%d %r value needs to be a float for datetime conversion' % \
									    (i, n))

					v = datetime.fromtimestamp(v)

				row[n] = v

			return row


		# Querying.
		# Todo: cursors
		def get(self, column, value):
			# Select individual entity wrapped row.
			column = baseStringClass._asString(column)
			if isinstance(value, str):
				value = baseStringClass._asString(value)

			return vmCurrentTask().sequence \
				(self._Entity(self, row = e)
				 for e in self._tableClass.selectBy \
					(**{column: value}))

		def get(self, column, value):
			return self.query({column: value})

		def all(self):
			# Return native object query set.
			return vmCurrentTask().sequence \
				(self._Entity(self, row = e)
				 for e in self._tableClass.select())

		def query(self, criteria):
			asString = baseStringClass._asString
			def asStringIf(v):
				if isinstance(v, str):
					return asString(v)

				return v

			criteria = dict((asString(name), asStringIf(value))
				            for (name, value) in criteria.items())

			return vmCurrentTask().sequence \
				(self._Entity(self, row = e)
				 for e in self._tableClass.selectBy(**criteria))


		# Migrations.
		# todo: emit underlying alter table sql commands to migrate the database.
		# todo: recreate the underlying sqlobject table class by deregistering
		# it with the class registry first.
		def addColumn(self, name, type):
			pass
		def removeColumn(self, name):
			pass

		def addColumns(self, *columns):
			# Add multiple columns, specified with 2-tuples (or sequences) of
			# name and type.
			pass
		def removeColumns(self, *columns):
			# Remove multiple columns by name.
			pass


		class _Entity(writeprotected):
			def __init__(self, table, row = None, values = None):
				self._table = table
				self._row = row
				self._values = values
				self.__public_members__ = table._writables

			# todo: do column set based on row information.
			# def __setattr__(self, name, value):
			# 	pass

			@property
			def id(self):
				return getattr(self._row, 'id', None)

			def __repr__(self):
				if self._row is None:
					return f'<{self._table._name} {self._values}>'

				return f'<{self._table._name} {self._row}>'

			def __getattr__(self, name):
				try: return object.__getattribute__(self, name)
				except AttributeError as e:
					try: return self.getValue(name)
					except KeyError:
						raise e

			@property
			def _columnNames(self):
				return (n for (n, c) in self._table._definition)

			@property
			def _dataItems(self):
				v = self._values
				if v is None:
					return ((n, getattr(self, n)) for n in self._columnNames)

				return ((n, v[n]) for n in self._columnNames)

			def getValue(self, name):
				if self._row is None:
					return self._values[name]

				if any(name == n for (n, c) in self._table._definition):
					return getattr(self._row, name)

				raise AttributeError(name)

			def _getColumnValues(self):
				return [v for (n, v) in self._dataItems]

			def _saveNoAccount(self):
				if self._row is None:
					self._row = self._table._tableClass(**self._values) # insert

				else:
					self._row.sync()

				return self


			def save(self):
				if self._table._database.noAccount:
					return self._saveNoAccount()

				def sizeOf(v):
					if isinstance(v, str):
						return len(v)

					return 4 # or 8?

				def insertAcct(size):
					acct = self._table._accountingTable(object = self._row.id,
						                                size = size)
					acct.sync()
					del self._table.size

					return acct

				def checkInsert(size, r):
					# Check for quota boundaries.
					m = self._table._database.size + size
					if m > r:
						raise IndexError('%s > %s' % (m, r))


				size = sum(sizeOf(v) for v in self._getColumnValues())

				if self._row is None:
					r = self._table._database._hard
					if r >= 0:
						checkInsert(size, r)

					self._row = self._table._tableClass(**self._values) # insert
					insertAcct(size)

				else:
					# Check for quota boundaries.
					r = self._table._database._hard
					if r >= 0:
						for acct in self._table._accountingTable.select(object = self._row.id):
							if acct.size != size:
								m = self._table._database.size + (size - acct.size)
								if m > r:
									raise IndexError('%s > %s' % (m, r))

								acct.size = size
								acct.sync()
								del self._table.size

							break
						else:
							# Catch not exist for upsert.
							checkInsert(size, r)
							insertAcct(size)

					self._row.sync()

				return self

	@classmethod
	def _InstallNative(self, core):
		path = configuration.AgentSystem.variable_database or 'components/services/database'
		core.newToolConfig(package = VariableDBNativeTool(core, self),
						   path = path)


# def table(name, **cfg):
#     return newClass(name, (sqlite.Object,) **value)

# used?
vardb = VariableDB._Open
createdb = VariableDB._Create

# table = VariableDB.Table

# vardb('basic', house = table('house', id = table.integer, guests = table.list)) \
#     .objects.house(id = 0, guests = [0]).sync()

def db(self, name, value, **kwd):
	from stuphos.language.document.interface import getContextEnvironment

	# debugOn()
	# Todo: the full object name in the yaml spec isn't used: just the last spec'd name.
	# This should be folded into another context environment variable.
	try: path = getContextEnvironment('document')
	except KeyError: pass
	else: name = '%s/%s' % ('/'.join(path), name)

	return vardb(name, **value)

def table(self, name, value, **kwd):
	def buildColumn(c):
		if len(c) != 1:
			raise ValueError('Column must have one key-value pair.')

		(name, type) = list(c.items())[0]
		c = VariableDBNativeTool._columns[type]._columnClass(name)
		return (c.name, c)

	name = name.replace('$', '_').replace('@', '_').replace('.', '_').replace('-', '_').replace('%', '_')

	assert not name.endswith('_acct'), NameError('Table name %r must not end in _acct' % name)
	assert isinstance(value, list)
	cols = [buildColumn(c) for c in value]
	return VariableDB.Table(name, cols)

installVariableDatabaseNativeTool = VariableDB._InstallNative



# Utilizes libdata.
# Todo: Load into world.entities!


# Todo: this function name says nothing about what it does.  In this case 'new' means 'load' from index.
def newWorld(stuphlib, base, index = None, cascade = False, verbose = False):
    if verbose: from sys import stdout as logging
    else: logging = None

    base = stuphlib.dblib.LocalIndex(base)
    if index:
        w = stuphlib.wldlib.World(base, stuphlib.dblib.isAbsIndex(index), logStream = logging)
        w.loadFromZoneIndex(index, cascade = cascade)
    else:
        w = stuphlib.wldlib.World(base, logStream = logging)
        w.loadFromZoneIndex(cascade = cascade)

    return w

class Database:
    # World & Adjunct Records.
    def __init__(self, stuphlib, worldModule, world_dir, zone_index = None, cascade = False, verbose = False):
        self.stuphlib = stuphlib
        self.worldModule = worldModule

        self.world_dir = world_dir
        self.zone_index = zone_index
        self.cascade = cascade
        self.verbose = verbose

        self.world = None

        # from game.misc import profile
        # self.boot = profile(filename = 'db.boot.prof')(self.boot)

    def __repr__(self):
        return '[DB %r]' % self.world

    def boot(self):
        self.world = newWorld(self.stuphlib,
                              self.world_dir,
                              self.zone_index,
                              self.cascade,
                              self.verbose)

        # Load into world entities!
        from sys import stderr as logOutput

        if self.verbose:
            print('Internalizing World...', file = logOutput)

            from stuphos.etc.tools.timing import Elapsed
            e = Elapsed()

        for z in self.world.zones:
            if self.verbose:
                logOutput.write('\rLoading %r...' % z)
                logOutput.flush()

            # import pdb; pdb.set_trace()
            zone = self.worldModule.zone.create(record = z)

            # Patch up.
            z.resetReady = zone.resetReady
            z.reset = zone.reset

            for r in z.rooms:
                try: room = zone.createRoom(r.vnum, zone = zone, record = r)
                except ValueError:
                    print('ERROR: Room %r already exists!' % r, file = logOutput)

            for i in z.items:
                try: item = zone.createItem(i.vnum, zone = zone, record = i)
                except ValueError:
                    print('ERROR: Item %r already exists!' % i, file = logOutput)

            for m in z.mobiles:
                try: mob = zone.createMobile(m.vnum, zone = zone, record = m)
                except ValueError:
                    print('ERROR: Mobile %r already exists!' % m, file = logOutput)

        if self.verbose:
            print(file = logOutput) # the newline
            print('...', e, 'Resolving Exits...', file = logOutput)

            e = Elapsed()

        self.resolve_exits()

        if self.verbose:
            print('...', e, file = logOutput)

    def resolve_exits(self):
        def resolveRoomExits(room):
            for exit in room.exits:
                # Dereference real room.
                try: r = self.worldModule.room(exit.roomlink)
                except ValueError:
                    # Mini-mud?
                    r = None

                exit.room = exit.destination = r


        self.worldModule.iterate_entities(self.worldModule.room, resolveRoomExits)

    def zoneUpdate(self, sec, usec):
        def resetZone(zone):
            if zone.resetReady():
                self.resetZone(zone)

        self.worldModule.iterate_entities(self.worldModule.zone, resetZone)

    def resetZone(self, zone):
        # Emit lifecycle events.
        # Cycle through zone commands.
        print('Resetting', repr(zone))
        zone.reset()

    def printMaster(self):
        for zone in self.worldModule.zone.table.values():
            print(zone)


    def resetWorld(self, *args, **kwd):
        # Only run if stuphmud.server.zones.controller.ZoneReset is a service.
        zreset = runtime[runtime.StuphMUD.Zone.Controller]
        if zreset is not None:
            if self.verbose:
                print('[System Database] Resetting World...')

            return zreset.resetWorld(*args, **kwd)



#!python
from common.subsystem.project.distribution import structured

CMRWebsite = ''

# todo: keywords, classifiers, fix author email, update version,
# get other attributes straight from the core.  determine license.

# fix using .pyc

@structured('MN') # .run
def setup(api):
    '''
    Westmetal Configuration::
        project = common.subsystem.project.distribution

    # todo: somehow tag this release as a 'cmr'
    MN(project$distribution):
        description: &description 'Application software.'
        summary: *description

        copyright: 2011-2019 All rights reserved.

        version: 0.0
        url: ''

        license: none
        classifiers: []
        keywords: []

        author: pypiContributor
        author-email: none

        include-package-data: true

        # todo: put this all under description structure.
        ignore-packages: true
        ignore-description: true

        document:
            indent($expression): indent
            report::

                by {project[author]}, {environment[copyright]}

                This package contains a Composition Milestone Report (CMR) for:
                    Name: {project[name]}
                    Version: {project[version]}

                For information about a CMR, see:
                    {description.website}

                {description.indented}

                About this project:

                    {long_description}


        customization($method):
            parameters: [appl, client, config] # , +args, ++kwd]
            code::
                try: appl.deleteConfiguration('packages')
                except KeyError: pass

                report = container.document.report.format

                with appl.distributionApi as api:
                    kwd = dict(appl.meta,
                               description = client(api),
                               long_description = appl['description'],
                               application = appl,
                               client = client,
                               project = appl.settings,
                               environment = appl.environment)

                    appl.modifyConfiguration \
                        ('long_description',
                         report(**kwd))

                         '''

    # XXX Don't want to be using io.here, want to be using another
    # folder target, especially because the script will end up using
    # this exported (distribution) package.
    #
    # This is bad because it generates .pyc, thereby polluting our
    # checksum.  I suppose this is ok for _any other project_, but
    # alas, bootstrap.
    #
    # Because this is NOT a distutils command, just try to find landmark.
    from common.subsystem.project.cmr import ProjectHash

    # folder = io.path(__file__).folder

    def findNewest():
        MONTHS = ['jan', 'feb', 'mar', 'apr', 'may', 'jun',
                  'jul', 'aug', 'sep', 'oct', 'nov', 'dec']

        from common.runtime.layer.etc import _daily_time_pattern_base as timestamp

        import re
        pattern = re.compile(r'wrlc-%s-r(\d+)' % timestamp)

        newest = None

        for folder in io.here.baseFilter('wrlc-*-r*'):
            a = folder.basename
            m = pattern.match(a)

            if m is not None:
                (month, day, year, revision) = m.groups()
                month = MONTHS.index(month.lower())
                day = int(day)
                year = int(year)
                revision = int(revision)

                encoded = '%04d%02d%02d%08d' % (month, day, year, revision)
                if newest and encoded < newest[0]:
                    continue

                newest = [encoded, folder]

        if newest is None:
            return io.here

        return newest[1]

    # Are there other ways to pass parameters to setup distributions?
    folder = system.module.os.environ.get('CMRTARGET') or findNewest()
    folder = io.path(folder)
    hash = ProjectHash.GenerateFromProjectFolder(folder)

    return synthetic(text = hash, website = CMRWebsite,
                     indented = indent(hash)) # XXX indenting signature line...


if __name__ == '__main__':
    import sys
    # import pdb; pdb.set_trace()
    if len(sys.argv) >= 2 and sys.argv[1] == '--dist':
        # Invoke traditional setuptools.
        del sys.argv[0]

        try: from setuptools import setup, find_packages
        except ImportError: from distutils.core import setup, find_packages

        SETUP_CONF = \
        dict (name = "MN",
              description = "Application software.",
              download_url = "",

              license = "None",
              platforms = ['OS-independent', 'Many'],

              include_package_data = True,

              keywords = [],

              classifiers = [])

        SETUP_CONF['version'] = '0.0'
        SETUP_CONF['url'] = ''

        SETUP_CONF['author'] = ''
        SETUP_CONF['author_email'] = ''

        SETUP_CONF['long_description'] = ''
        SETUP_CONF['packages'] = find_packages()

        setup(**SETUP_CONF)

    else:
        setup.main()



# Automatic service configuration for zone modules.
#
from stuphos.runtime.registry import getObject as getRegistryObject, RegistryNotInstalled
from ..etc import isYesValue, capitalize, getSystemException, logException

from .config import getStandardPath, getParentPath, loadConfig
from .config import getStandardPath, getParentPath, loadConfig

def getEnablingSection(name):
    # A shorter, sweeter version.
    from stuphos import getSection
    from ..etc import isYesValue

    cfg = getSection(name)
    def isEnabled(option):
        return isYesValue(cfg.get(option))

    return isEnabled

class AutoServices:
    CONFIG_SECTION_NAME = 'Services'

    @classmethod
    def ParentPath(self, base, *parts):
        return self.JoinPath(getParentPath(base), *parts)

    @classmethod
    def JoinPath(self, *parts):
        return getStandardPath(*parts)

    def __init__(self, setup_deferred = False):
        self.reset()

        # Do the actual object setup during load because maybe the
        # (configuration) objects aren't around yet.
        self.setup_deferred = setup_deferred
        if not setup_deferred:
            self.setup()

    def setup(self):
        try:
            configObj = getRegistryObject(self.CONFIG_OBJECT_NAME,
                                          create = lambda:loadConfig \
                                          (self.CONFIG_FILE))

            self.section = configObj.getSection(self.CONFIG_SECTION_NAME)
            self.configObj = configObj
        except RegistryNotInstalled:
            self.section = None

    def getConfigObject(self):
        return self.configObj
    def getConfigSection(self):
        return self.section
    def getConfig(self, name, section = None):
        return self.getConfigObject().get(name, section = self.CONFIG_OBJECT_NAME
                                          if section is None else section)

    def reloadConfig(self):
        # todo: use method-based api..
        del runtime[self.CONFIG_OBJECT_NAME]
        self.setup()

    def reset(self):
        self.loaded = set()

    class ServiceConf:
        def __init__(self, name, value):
            self.name = name
            self.value = value

    def getServiceLoaderName(self, name):
        return 'load' + ''.join(capitalize(part) for part in name.split('-'))

    def findServiceLoader(self, conf):
        methodName = self.getServiceLoaderName(conf.name)
        return getattr(self, methodName, None)

    def iterateConfServices(self):
        if self.section:
            for opt in self.section.options():
                conf = self.ServiceConf(opt, self.section.get(opt))
                function = self.findServiceLoader(conf)

                if callable(function):
                    yield (conf, function)

    def shouldLoad(self, conf):
        try:
            if conf.name not in self.loaded:
                return isYesValue(conf.value)
        finally:
            self.loaded.add(conf.name)

    def __call__(self):
        if self.setup_deferred:
            self.setup()

        for (conf, load) in self.iterateConfServices():
            if self.shouldLoad(conf):
                # debugOn()
                try: load(conf)
                except SyntaxError as e:
                    print('There was a syntax error in module configuration for service:', conf.name)
                    #'args', 'filename', 'lineno', 'message', 'msg', 'offset', 'print_file_and_line', 'text']
                    print('   ', e.filename, '(%d:%d)' % (e.lineno, e.offset))
                except: self.error(*getSystemException())

        self.postOn(self.section)

    def error(self, etype, evalue, tb):
        # Non-fatal.
        logException(etype, evalue, tb, traceback = True)

    def postOn(self, cfg):
        # Called after loading services.  Customize post-load.
        pass



"CircleMUD shop database file interpreter."
from .dblib import RecordReader

class ShopReader(RecordReader):
	VERSION3_TAG="v3.0"

	# Read all records in file
	def parseFile(self):
		'Override'
		line = self.readLineNoEOF("Shop Version Line (First)")

		if self.VERSION3_TAG in line:
			self.v3format = True

		self.startOfRecords()

		while 1:
			line = self.readLineNoEOF("Record VNR")

			c = line[0]
			if c == '$':
				break

			assert c == '#', ('Line %d: ' % self.lineno) + repr("[%s != '#'] %s" % (c, line[1:]))

			# This hack simply ignores all non-digits and goes from there.
			vno = int (''.join([n for n in line if n.isdigit()]))

			self.newRecord(vno)
			self.parseRecord()

		self.endOfRecords()

	def readBuyType(self):
		"Returns None on malformat; empty-string on end-of-list indicator (-1); otherwise a 2-tuple."
		try:
			line=self.readLine().strip()

			i=line.find(';')
			line=(i>=0) and line[:i] or line[:-1]

			parts=line.split(' ')
			num=int(parts[0])

			return (num==-1) and '' or (num, parts[1])

		except:
			return None

	def readBuyTypeList(self):
		if not getattr(self, 'v3format'):
			return self.readNumberList()

		buyTypes=[];append=buyTypes.append

		while True:
			bt=self.readBuyType()
			if not bt:
				break

			append(bt)

		return buyTypes

	# read_shop_message validates %-formatters: we'll skip this.
	readShopMessage = RecordReader.tildeString

	def parseRecord(self):
		"Load record area into logical Python object."

		# boot_the_shops sets the 'zone' member: we'll skip this.

		self.recordField('producing',     self.readNumberList())

		self.recordField('buy-profit',    self.readFloat())
		self.recordField('sell-profit',   self.readFloat())

		self.recordField('buy-type',      self.readBuyTypeList())

		self.recordField('no-such-item1', self.readShopMessage())
		self.recordField('no-such-item2', self.readShopMessage())
		self.recordField('do-not-buy',    self.readShopMessage())
		self.recordField('missing-cash1', self.readShopMessage())
		self.recordField('missing-cash2', self.readShopMessage())
		self.recordField('message-buy',   self.readShopMessage())
		self.recordField('message-sell',  self.readShopMessage())

		self.recordField('temper',        self.readInteger())
		self.recordField('bitvector',     self.readInteger())

		self.recordField('shopkeeper',    self.readInteger())
		self.recordField('trade-with',    self.readInteger())

		self.recordField('rooms',         self.readNumberList())

		self.recordField('open-1',        self.readInteger())
		self.recordField('open-2',        self.readInteger())
		self.recordField('close-1',       self.readInteger())
		self.recordField('close-2',       self.readInteger())


if __name__=='__main__':
	from glob import glob
	from sys import argv
	from os.path import join

	argv = argv[1:]
	if argv:
		x = [ShopReader(n)._() for n in glob(join(argv[0], '*.shp'))]
